import math

from .. import utils
from .feature_data import *


def robust_cholesky(matrix: torch.Tensor) -> torch.Tensor:
    """
    Implements a Cholesky decomposition.
    If the Cholesky decomposition fails, it is retried with increasing added jitter on the diagonal.
    :param matrix: Symmetric positive semi-definite matrix to factorize.
    :return: Approximate cholesky factor L such that (approximately) LL^T = matrix
    """
    eps = 1e-5 * matrix.trace() / matrix.shape[-1]
    L = None
    for i in range(10):
        try:
            L = torch.linalg.cholesky(matrix)
            break
        except RuntimeError:
            print('Increasing jitter for Cholesky decomposition', flush=True)
            matrix += eps * torch.eye(matrix.shape[-1], device=matrix.device, dtype=matrix.dtype)
            eps *= 2
    if L is None:
        raise RuntimeError('Could not Cholesky decompose the matrix')
    return L


def robust_cholesky_inv(matrix: torch.Tensor) -> torch.Tensor:
    """
    :param matrix: Symmetric positive semi-definite matrix.
    :return: A matrix A such that (approximately) matrix^{-1} = A^T A, where A = L^{-1} for the Cholesky factor L.
    """
    # returns a matrix A such that matrix^{-1} = A^T A
    L = robust_cholesky(matrix)
    # there might be a better solution, but I found no direct triangular inversion function
    return L.inverse()


class DataTransform:
    """
    Abstract base class for representing functions that transform FeatureData objects into other FeatureData objects,
    usually by applying the same function to each sample in the FeatureData object.
    This is typically used for (parts of) feature maps, except that it is more general than a feature map
    since it does not need to allow the computation of (inner products of) vector-valued features.
    For example, in a feature map phi(x) = f(g(x)), f is also a feature map, but g only needs to be a DataTransform.
    """
    def __call__(self, feature_data: FeatureData, idxs: Optional[Indexes] = None) -> FeatureData:
        """
        Method to apply the transformation to (a subset of) the given feature data.
        Subclasses should not override this method, but override forward() instead.
        :param feature_data: FeatureData to apply this transformation to.
        :param idxs: An object to index the feature data with, or None if all of the feature data should be used.
        :return: Returns a FeatureData object.
        """
        idxs = Indexes(feature_data.get_n_samples(), idxs)
        pieces = [self.forward(sub_data, sub_idxs)
                  for sub_idxs, sub_data in feature_data.iterate(idxs)]
        return ConcatFeatureData(pieces) if len(pieces) != 1 else pieces[0]

    def forward(self, feature_data: FeatureData, idxs: Indexes) -> FeatureData:
        """
        Internal function for subclasses to override.
        :param feature_data: FeatureData to apply the transformation to.
        :param idxs: Indexes object corresponding to the subset of the feature data that should be processed.
        :return: Transformed FeatureData.
        """
        raise NotImplementedError()


class FeatureMap(DataTransform):
    """
    Abstract base class for representing feature maps and their corresponding kernel.
    For kernels with infinite-dimensional feature space,
    the corresponding sub-classes may only allow evaluation of the kernel and not of the feature map.
    """
    def __init__(self, n_features: int, allow_precompute_features: bool = True):
        """
        :param n_features: Feature space dimension of the feature map.
        Should be -1 if the feature space dimension is infinite.
        :param allow_precompute_features: Specifies whether the default behavior of precompute()
        should be to simply precompute the feature matrix.
        This might not be desirable for kernels with high feature-space dimension
        where more efficient kernel evaluation methods than the inner product between feature vectors exist.
        """
        self.n_features = n_features
        self.allow_precompute_features = allow_precompute_features

    def get_n_features(self):
        """
        :return: Returns the feature space dimension.
        """
        return self.n_features

    def forward(self, feature_data: FeatureData, idxs: Indexes) -> FeatureData:
        """
        Implements the forward() method from DataTransform,
        which in this case computes the feature matrix wrapped in TensorFeatureData.
        This method can only be called if self.get_feature_matrix_impl_() is implemented.
        :param feature_data: Input data to the feature map.
        :param idxs: Indexes to index feature_data with.
        :return: Returns a TensorFeatureData object containing the feature matrix.
        """
        return TensorFeatureData(self.get_feature_matrix(feature_data, idxs))

    def precompute(self, feature_data: FeatureData, idxs: Optional[Indexes] = None) -> Tuple['FeatureMap', FeatureData]:
        """
        Returns a tuple (fm, fd) such that the feature map fm on the feature data fd
        behaves identically to this feature map on feature_data, in terms of kernel matrix values.
        For example, this method may transform (phi, x) to (id, phi(x)), if self = phi and x = feature_data.
        The returned tuple should be at least as fast to evaluate as before.
        Subclasses should not override precompute() but precompute_soft_().
        :param feature_data: Input to this feature map.
        :param idxs: Indexes to index feature_data with, or None if the full data should be used.
        :return: Returns an equivalent tuple of FeatureMap and FeatureData.
        """
        idxs = Indexes(feature_data.get_n_samples(), idxs)
        if self.allow_precompute_features:
            # simply compute the feature matrix and apply an identity feature map to it
            return IdentityFeatureMap(n_features=self.n_features), self(feature_data, idxs)
        else:
            # use another precomputation method that can be overridden by a subclass.
            results = [self.precompute_soft_(sub_data, sub_idxs) for sub_idxs, sub_data in feature_data.iterate(idxs)]
            if len(results) == 1:
                return results[0]
            # we do not call .simplify() on the feature data here
            # since precompute() might be called recursively on smaller parts of the data,
            # and we do not want to simplify() multiple times but only once after precompute()
            return results[0][0], ConcatFeatureData([r[1] for r in results])

    def precompute_soft_(self, feature_data: FeatureData, idxs: Indexes) -> Tuple['FeatureMap', FeatureData]:
        """
        Internal method to precompute the feature map
        in the case that precomputing the whole feature matrix is not allowed.
        This method can be overridden by subclasses if needed.
        By default, it does not perform any precomputations and just returns (self, feature_data[idxs]).
        :param feature_data: Feature data to apply this feature map to.
        :param idxs: Indexes to index the feature data with.
        :return: Potentially precomputed tuple (FeatureMap, FeatureData).
        """
        # we do not call .simplify() on the feature data here
        # since precompute() might be called recursively on smaller parts of the data,
        # and we do not want to simplify() multiple times but only once after precompute()
        return self, feature_data[idxs]

    def posterior(self, feature_data: FeatureData, sigma: float, allow_kernel_space_posterior: bool = True) \
            -> 'FeatureMap':
        """
        Returns a feature map that represents the Gaussian Process posterior kernel after observing feature_data,
        if the noise variance is sigma^2.
        :param feature_data: Feature data to condition the Gaussian Process with.
        :param sigma: Noise standard deviation for the Gaussian Process.
        :param allow_kernel_space_posterior: Whether the method is allowed to use the kernel-space posterior formula
        if it is deemed more efficient but not strictly necessary. The kernel-space posterior feature map returned
        will not allow a computation of the feature matrix, which could be detrimental
        for methods that want to operate in feature space.
        :return: Returns a feature map that represents the Gaussian Process posterior kernel after observing feature_data,
        if the noise variance is sigma^2.
        """
        if self.n_features < 0 or (allow_kernel_space_posterior and self.n_features > max(1024, 3 * len(feature_data))):
            # compute the posterior in kernel space
            return KernelSpacePosteriorFeatureMap(feature_map=self, cond_data=feature_data, sigma=sigma)

        feature_matrix = self.get_feature_matrix(feature_data)
        eye = torch.eye(self.n_features, device=feature_matrix.device, dtype=feature_matrix.dtype)
        cov_matrix = feature_matrix.t().matmul(feature_matrix) + (sigma ** 2) * eye

        return SequentialFeatureMap(LinearFeatureMap(sigma * robust_cholesky_inv(cov_matrix).t()), [self])

    def get_feature_matrix(self, feature_data: FeatureData, idxs: Optional[Indexes] = None) -> torch.Tensor:
        """
        Returns the feature matrix obtained by applying this feature map to feature_data[idxs].
        This method can only be used if get_feature_matrix_impl_() is implemented,
        in particular it cannot be used if the feature space dimension is infinite.
        Subclasses should not override this method but get_feature_matrix_impl_() instead.
        :param feature_data: FeatureData to apply this feature map to.
        :param idxs: Indexes to index feature_data with, or None to use the full feature_data.
        :return: Feature matrix as torch.Tensor of shape n_samples x n_features
        """
        idxs = Indexes(feature_data.get_n_samples(), idxs)
        return torch_cat([self.get_feature_matrix_impl_(sub_data, sub_idxs)
                          for sub_idxs, sub_data in feature_data.iterate(idxs)], dim=-2)

    def get_kernel_matrix(self, feature_data_1: FeatureData, feature_data_2: FeatureData,
                          idxs_1: Optional[Indexes] = None, idxs_2: Optional[Indexes] = None) -> torch.Tensor:
        """
        Returns the kernel matrix k(feature_data_1[idxs_1], feature_data_2[idxs_2]),
        where k is the kernel corresponding to the current feature map.
        Subclasses should not override this method but get_kernel_matrix_impl_() instead.
        :param feature_data_1: First feature data.
        :param feature_data_2: Second feature data.
        :param idxs_1: Indexes for first feature data, or None.
        :param idxs_2: Indexes for second feature data, or None.
        :return: Kernel matrix as torch.Tensor of shape n_samples_1 x n_samples_2
        """
        idxs_1 = Indexes(feature_data_1.get_n_samples(), idxs_1)
        idxs_2 = Indexes(feature_data_2.get_n_samples(), idxs_2)
        return torch_cat([torch_cat([self.get_kernel_matrix_impl_(sub_data_1, sub_data_2, sub_idxs_1, sub_idxs_2)
                                     for sub_idxs_2, sub_data_2 in feature_data_2.iterate(idxs_2)], dim=-1)
                          for sub_idxs_1, sub_data_1 in feature_data_1.iterate(idxs_1)], dim=-2)

    def get_kernel_matrix_diag(self, feature_data: FeatureData, idxs: Optional[Indexes] = None) -> torch.Tensor:
        """
        Returns the diagonal of the kernel matrix k(feature_data[idxs], feature_data[idxs]),
        where k is the kernel corresponding to the current feature map.
        Subclasses should not override this method but get_kernel_matrix_diag_impl_() instead.
        :param feature_data: Feature data.
        :param idxs: Indexes to index the first feature data, or None.
        :return: Diagonal of Kernel matrix as torch.Tensor of shape [n_samples]
        """
        idxs = Indexes(feature_data.get_n_samples(), idxs)
        return torch_cat([self.get_kernel_matrix_diag_impl_(sub_data, sub_idxs)
                          for sub_idxs, sub_data in feature_data.iterate(idxs)], dim=-1)

    def get_feature_matrix_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        """
        Internal method that should be overridden by subclasses to compute the feature matrix.
        :param feature_data: Feature data to apply this feature map to.
        :param idxs: Indexes to index feature_data.
        :return: Feature matrix as torch.tensor of shape n_samples x n_features
        """
        raise NotImplementedError()

    def get_kernel_matrix_impl_(self, feature_data_1: FeatureData, feature_data_2: FeatureData,
                                idxs_1: Indexes, idxs_2: Indexes) -> torch.Tensor:
        """
        Internal method to compute the kernel matrix.
        By default, this method uses get_feature_matrix()
        and then computes a matrix-matrix product of two feature matrices.
        Subclasses should override this method if the kernel matrix should be computed in another way.
        :param feature_data_1: First feature data.
        :param feature_data_2: Second feature data.
        :param idxs_1: Indexes for the first feature data.
        :param idxs_2: Indexes for the second feature data.
        :return: Kernel matrix as torch.Tensor of shape n_samples_1 x n_samples_2
        """
        return self.get_feature_matrix(feature_data_1, idxs_1).matmul(
            self.get_feature_matrix(feature_data_2, idxs_2).t())

    def get_kernel_matrix_diag_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        """
        Internal method to compute the diagonal of the kernel matrix.
        By default, this method uses the squared norm of the feature vectors obtained by get_feature_matrix().
        Subclasses should override this method if another computation should be used.
        :param feature_data: Feature data.
        :param idxs: Indexes to index the feature data.
        :return: Kernel matrix diagonal as torch.Tensor of shape [n_samples]
        """
        return (self.get_feature_matrix(feature_data, idxs) ** 2).sum(dim=-1)

    def sketch(self, n_features: int, **config) -> 'FeatureMap':
        """
        Apply sketching (a.k.a. random projections) to this feature map,
        returning a feature map with (lower) feature-space dimension given by n_features.
        This method may not be implemented by all subclasses.
        :param n_features: Number of target features for the sketched feature map.
        :param config: This can be used to pass other parameters to individual sketching operations.
        :return: Returns a sketched feature map. Note that this does not modify any data yet,
        for this the sketched feature map needs to be precomputed on the data.
        """
        raise NotImplementedError()


class IdentityFeatureMap(FeatureMap):
    """
    This class represents the identity feature map phi(x) = x, and the linear kernel k(x, y) = x^T y.
    """
    def __init__(self, n_features: int):
        """
        :param n_features: Dimension of the inputs.
        """
        super().__init__(n_features=n_features)

    def get_feature_matrix_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        return feature_data.get_tensor(idxs)

    def sketch(self, n_features: int, **config) -> 'FeatureMap':
        # Gaussian sketch
        matrix = torch.randn(self.n_features, n_features)
        if config.get('sketch_norm', False):
            # it is possible to normalize the Gaussian vectors, which can make a difference for small input dimensions.
            matrix /= (matrix ** 2).sum(dim=0, keepdim=True).sqrt() / math.sqrt(self.n_features)
        matrix /= math.sqrt(n_features)
        return SequentialFeatureMap(LinearFeatureMap(matrix), [self])


class ReLUNTKFeatureMap(FeatureMap):
    """
    This feature map represents the Neural Tangent Kernel (Jacot et al., 2018) corresponding to a ReLU NN
    in Neural Tangent Parameterization. We implement the form of Lee et al. (2019), with factors sigma_w and sigma_b,
    and where the biases are initialized from N(0, 1).
    """
    # following SM 3 and 5 in
    # https://proceedings.neurips.cc/paper/2019/hash/0d1a9651497a38d8b1c3871c84528bd4-Abstract.html
    def __init__(self, n_layers=3, sigma_w_sq=2.0, sigma_b_sq=0.0):
        """
        :param n_layers: Number of layers of the corresponding NN
        :param sigma_w_sq: sigma_w**2 in the notation of Lee et al. (2019)
        :param sigma_b_sq: sigma_b**2 in the notation of Lee et al. (2019)
        """
        super().__init__(n_features=-1, allow_precompute_features=False)
        self.n_layers = n_layers
        self.sigma_w_sq = sigma_w_sq
        self.sigma_b_sq = sigma_b_sq

    def t_and_tdot_(self, a: torch.Tensor, b: torch.Tensor, d: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        a_sqrt = a.sqrt()
        d_sqrt = d.sqrt()
        cos_theta = torch.clip(b / (a_sqrt * d_sqrt + 1e-30), min=-1.0, max=1.0)
        theta = torch.arccos(cos_theta)
        t = 1 / (2 * math.pi) * a_sqrt * d_sqrt * ((1 - cos_theta ** 2).sqrt() + (math.pi - theta) * cos_theta)
        tdot = 1 / (2 * math.pi) * (math.pi - theta)
        return t, tdot

    def diag_prop_(self, diag: torch.Tensor) -> torch.Tensor:
        # evaluate (k_l(x, x))_{x \in X} from diag = (k_{l-1}(x, x))_{x \in X}
        t = 0.5 * diag
        return self.sigma_w_sq * t + self.sigma_b_sq

    def get_kernel_matrix_impl_(self, feature_data_1: FeatureData, feature_data_2: FeatureData,
                                idxs_1: Indexes, idxs_2: Indexes) -> torch.Tensor:
        feature_mat_1 = feature_data_1.get_tensor(idxs_1)
        feature_mat_2 = feature_data_2.get_tensor(idxs_2)
        kernel_mat = feature_mat_1.matmul(feature_mat_2.t())
        d_in = feature_mat_1.shape[1]
        diag_1 = self.sigma_w_sq / d_in * (feature_mat_1 ** 2).sum(dim=1) + self.sigma_b_sq
        diag_2 = self.sigma_w_sq / d_in * (feature_mat_2 ** 2).sum(dim=1) + self.sigma_b_sq
        nngp_mat = self.sigma_w_sq / d_in * kernel_mat + self.sigma_b_sq
        ntk_mat = nngp_mat
        for i in range(self.n_layers - 1):
            t, tdot = self.t_and_tdot_(diag_1[:, None], nngp_mat, diag_2[None, :])
            nngp_mat = self.sigma_w_sq * t + self.sigma_b_sq
            ntk_mat = nngp_mat + self.sigma_w_sq * ntk_mat * tdot
            diag_1 = self.diag_prop_(diag_1)
            diag_2 = self.diag_prop_(diag_2)
        return ntk_mat

    def get_kernel_matrix_diag_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        feature_mat = feature_data.get_tensor(idxs)
        d_in = feature_mat.shape[1]
        nngp_diag = self.sigma_w_sq / d_in * (feature_mat ** 2).sum(dim=1) + self.sigma_b_sq
        ntk_diag = nngp_diag
        for i in range(self.n_layers - 1):
            nngp_diag = self.diag_prop_(nngp_diag)
            ntk_diag = nngp_diag + self.sigma_w_sq * 0.5 * ntk_diag
        return ntk_diag


class ReLUNNGPFeatureMap(FeatureMap):
    """
    This feature map represents the NNGP Kernel (Jacot et al., 2018) corresponding to a ReLU NN
    in Neural Tangent Parameterization. We implement the form of Lee et al. (2019), with factors sigma_w and sigma_b,
    and where the biases are initialized from N(0, 1).
    """
    def __init__(self, n_layers=3, sigma_w_sq=2.0, sigma_b_sq=0.0):
        """
        :param n_layers: Number of layers of the corresponding NN
        :param sigma_w_sq: sigma_w**2 in the notation of Lee et al. (2019)
        :param sigma_b_sq: sigma_b**2 in the notation of Lee et al. (2019)
        """
        super().__init__(n_features=-1, allow_precompute_features=False)
        self.n_layers = n_layers
        self.sigma_w_sq = sigma_w_sq
        self.sigma_b_sq = sigma_b_sq

    def get_t_(self, a: torch.Tensor, b: torch.Tensor, d: torch.Tensor) -> torch.Tensor:
        a_sqrt = a.sqrt()
        d_sqrt = d.sqrt()
        cos_theta = torch.clip(b / (a_sqrt * d_sqrt + 1e-30), min=-1.0, max=1.0)
        theta = torch.arccos(cos_theta)
        t = 1 / (2 * math.pi) * a_sqrt * d_sqrt * ((1 - cos_theta ** 2).sqrt() + (math.pi - theta) * cos_theta)
        return t

    def diag_prop_(self, diag: torch.Tensor) -> torch.Tensor:
        # evaluate (k_l(x, x))_{x \in X} from diag = (k_{l-1}(x, x))_{x \in X}
        t = 0.5 * diag
        return self.sigma_w_sq * t + self.sigma_b_sq

    def get_kernel_matrix_impl_(self, feature_data_1: FeatureData, feature_data_2: FeatureData,
                                idxs_1: Indexes, idxs_2: Indexes) -> torch.Tensor:
        feature_mat_1 = feature_data_1.get_tensor(idxs_1)
        feature_mat_2 = feature_data_2.get_tensor(idxs_2)
        kernel_mat = feature_mat_1.matmul(feature_mat_2.t())
        d_in = feature_mat_1.shape[1]
        diag_1 = self.sigma_w_sq / d_in * (feature_mat_1 ** 2).sum(dim=1) + self.sigma_b_sq
        diag_2 = self.sigma_w_sq / d_in * (feature_mat_2 ** 2).sum(dim=1) + self.sigma_b_sq
        nngp_mat = self.sigma_w_sq / d_in * kernel_mat + self.sigma_b_sq
        for i in range(self.n_layers - 1):
            t = self.get_t_(diag_1[:, None], nngp_mat, diag_2[None, :])
            nngp_mat = self.sigma_w_sq * t + self.sigma_b_sq
            diag_1 = self.diag_prop_(diag_1)
            diag_2 = self.diag_prop_(diag_2)
        return nngp_mat

    def get_kernel_matrix_diag_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        feature_mat = feature_data.get_tensor(idxs)
        d_in = feature_mat.shape[1]
        nngp_diag = self.sigma_w_sq / d_in * (feature_mat ** 2).sum(dim=1) + self.sigma_b_sq
        for i in range(self.n_layers - 1):
            nngp_diag = self.diag_prop_(nngp_diag)
        return nngp_diag


class LaplaceKernelFeatureMap(FeatureMap):
    """
    Implements the Laplace kernel, k(x, y) = exp(-scale*||x-y||).
    """
    def __init__(self, scale: float = 1.0):
        """
        :param scale: Scale parameter of the Laplace kernel. Larger scale yields a narrower kernel.
        """
        super().__init__(n_features=-1, allow_precompute_features=False)
        self.scale = scale

    def get_kernel_matrix_impl_(self, feature_data_1: FeatureData, feature_data_2: FeatureData,
                                idxs_1: Indexes, idxs_2: Indexes) -> torch.Tensor:
        feature_mat_1 = feature_data_1.get_tensor(idxs_1)
        feature_mat_2 = feature_data_2.get_tensor(idxs_2)

        sq_dist_mat = (feature_mat_1**2).sum(dim=-1)[:, None] + (feature_mat_2**2).sum(dim=-1)[None, :] \
                   - 2 * feature_mat_1 @ feature_mat_2.t()
        return torch.exp(-self.scale*torch.sqrt(sq_dist_mat))

    def get_kernel_matrix_diag_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        feature_mat = feature_data.get_tensor(idxs)
        return torch.ones_like(feature_mat[:, 0])


class KernelSpacePosteriorFeatureMap(FeatureMap):
    """
    This internal class represents the posterior kernel of a GP after observing data.
    This is used internally by FeatureMap.posterior().
    """
    def __init__(self, feature_map: FeatureMap, cond_data: FeatureData, sigma: float):
        """
        :param feature_map: Prior feature map.
        :param cond_data: Data that the GP is conditioned on.
        :param sigma: Noise standard deviation of the GP.
        """
        super().__init__(n_features=-1, allow_precompute_features=False)
        mat = feature_map.get_kernel_matrix(cond_data, cond_data)
        # (K + sigma*I)^{-1} = L_inv.t() @ L_inv
        eye = torch.eye(mat.shape[0], device=mat.device, dtype=mat.dtype)
        self.L_inv = robust_cholesky_inv(mat + sigma**2 * eye)
        self.inv_mat = self.L_inv.t() @ self.L_inv
        self.feature_map = feature_map
        self.cond_data = cond_data
        self.sigma = sigma

    def precompute_soft_(self, feature_data: FeatureData, idxs: Indexes) -> Tuple['FeatureMap', FeatureData]:
        # Precomputation works as follows: The posterior kernel k_post is given by
        # k_post(x_1, x_2) = k(x_1, x_2) - k(x_1, X) (k(X, X) + sigma^2 I)^{-1} k(X, x_2) = k(x_1, x_2) - z_1^T z_2,
        # where z_i = (k(X, X) + sigma^2 I)^{-1/2} k(X, x_i).
        # Precomputation precomputes the z_i and appends them to the x_i via ListFeatureData.
        tensor = self.feature_map.get_kernel_matrix(feature_data, self.cond_data, idxs) @ self.L_inv.t()
        return PrecomputedKernelSpacePosteriorFeatureMap(self.feature_map), \
               ListFeatureData([feature_data[idxs].simplify(), TensorFeatureData(tensor)])

    def get_kernel_matrix_impl_(self, feature_data_1: FeatureData, feature_data_2: FeatureData,
                                idxs_1: Indexes, idxs_2: Indexes) -> torch.Tensor:
        prior_mat = self.feature_map.get_kernel_matrix(feature_data_1, feature_data_2, idxs_1, idxs_2)
        mat_1 = self.feature_map.get_kernel_matrix(feature_data_1, self.cond_data, idxs_1)
        mat_2 = self.feature_map.get_kernel_matrix(feature_data_2, self.cond_data, idxs_2)
        return prior_mat - mat_1 @ self.inv_mat @ mat_2.t()

    def get_kernel_matrix_diag_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        prior_diag = self.feature_map.get_kernel_matrix_diag(feature_data, idxs)
        mat = self.feature_map.get_kernel_matrix(feature_data, self.cond_data, idxs)
        return prior_diag - ((mat @ self.L_inv.t()) ** 2).sum(dim=1)


class PrecomputedKernelSpacePosteriorFeatureMap(FeatureMap):
    """
    Internal class arising from precomputing on a KernelSpacePosteriorFeatureMap.
    """
    def __init__(self, feature_map: FeatureMap):
        """
        :param feature_map: Prior feature map, which is applied to the non-precomputed part of the data.
        """
        super().__init__(n_features=-1, allow_precompute_features=False)
        self.feature_map = feature_map

    def get_kernel_matrix_impl_(self, feature_data_1: FeatureData, feature_data_2: FeatureData,
                                idxs_1: Indexes, idxs_2: Indexes) -> torch.Tensor:
        if not isinstance(feature_data_1, ListFeatureData):
            raise ValueError(f'feature_data_1 must be of type ListFeatureData, but is of type {type(feature_data_1)}')
        if not isinstance(feature_data_2, ListFeatureData):
            raise ValueError(f'feature_data_2 must be of type ListFeatureData, but is of type {type(feature_data_2)}')
        prior_mat = self.feature_map.get_kernel_matrix(feature_data_1.feature_data_list[0],
                                                       feature_data_2.feature_data_list[0], idxs_1, idxs_2)
        tensor_1 = feature_data_1.feature_data_list[1].get_tensor(idxs_1)
        tensor_2 = feature_data_2.feature_data_list[1].get_tensor(idxs_2)
        return prior_mat - tensor_1 @ tensor_2.t()

    def get_kernel_matrix_diag_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        prior_diag = self.feature_map.get_kernel_matrix_diag(feature_data.feature_data_list[0], idxs)
        tensor = feature_data.feature_data_list[1].get_tensor(idxs)
        return prior_diag - (tensor ** 2).sum(dim=1)


class ScaledFeatureMap(FeatureMap):
    """
    Represents a scaled feature map factor * feature_map.
    """
    def __init__(self, feature_map: FeatureMap, factor: float):
        """
        :param feature_map: Feature map to scale.
        :param factor: Scaling factor.
        """
        super().__init__(n_features=feature_map.get_n_features(),
                         allow_precompute_features=feature_map.allow_precompute_features)
        self.feature_map = feature_map
        self.factor = factor

    def precompute_soft_(self, feature_data: FeatureData, idxs: Indexes) -> Tuple['FeatureMap', FeatureData]:
        fm, fd = self.feature_map.precompute(feature_data, idxs)
        return ScaledFeatureMap(fm, self.factor), fd

    def get_feature_matrix_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        return self.factor * self.feature_map.get_feature_matrix(feature_data, idxs)

    def get_kernel_matrix_impl_(self, feature_data_1: FeatureData, feature_data_2: FeatureData,
                                idxs_1: Indexes, idxs_2: Indexes) -> torch.Tensor:
        return self.factor ** 2 * self.feature_map.get_kernel_matrix(feature_data_1, feature_data_2, idxs_1, idxs_2)

    def get_kernel_matrix_diag_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        return self.factor ** 2 * self.feature_map.get_kernel_matrix_diag(feature_data, idxs)

    def sketch(self, n_features: int, **config) -> 'FeatureMap':
        return ScaledFeatureMap(self.feature_map.sketch(n_features, **config), self.factor)


class SumFeatureMap(FeatureMap):
    """
    Represents a feature map corresponding to the sum of multiple kernels.
    """
    def __init__(self, feature_maps: List[FeatureMap]):
        """
        :param feature_maps: Feature maps whose kernels should be added.
        """
        feature_counts = [fm.get_n_features() for fm in feature_maps]
        super().__init__(n_features=sum(feature_counts) if all([n >= 0 for n in feature_counts]) else -1,
                         allow_precompute_features=all([fm.allow_precompute_features for fm in feature_maps]))
        self.feature_maps = feature_maps

    def precompute_soft_(self, feature_data: FeatureData, idxs: Indexes) -> Tuple['FeatureMap', FeatureData]:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        results = [fm.precompute(fd, idxs) for fm, fd in zip(self.feature_maps, feature_data.feature_data_list)]
        return SumFeatureMap([r[0] for r in results]), ListFeatureData([r[1] for r in results])

    def get_feature_matrix_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        return torch_cat(
            [fm.get_feature_matrix(fd, idxs) for fm, fd in zip(self.feature_maps, feature_data.feature_data_list)],
            dim=-1)

    def get_kernel_matrix_impl_(self, feature_data_1: FeatureData, feature_data_2: FeatureData,
                                idxs_1: Indexes, idxs_2: Indexes) -> torch.Tensor:
        if not isinstance(feature_data_1, ListFeatureData):
            raise ValueError(f'feature_data_1 must be of type ListFeatureData, but is of type {type(feature_data_1)}')
        if not isinstance(feature_data_2, ListFeatureData):
            raise ValueError(f'feature_data_2 must be of type ListFeatureData, but is of type {type(feature_data_2)}')
        return sum([fm.get_kernel_matrix(fd1, fd2, idxs_1, idxs_2)
                    for fm, fd1, fd2 in zip(self.feature_maps, feature_data_1.feature_data_list,
                                            feature_data_2.feature_data_list)])

    def get_kernel_matrix_diag_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        return sum(
            [fm.get_kernel_matrix_diag(fd, idxs) for fm, fd in zip(self.feature_maps, feature_data.feature_data_list)])

    def sketch(self, n_features: int, **config) -> 'FeatureMap':
        return ElementwiseSumFeatureMap([fm.sketch(n_features, **config) for fm in self.feature_maps])


class ProductFeatureMap(FeatureMap):
    """
    Feature map corresponding to the product of multiple kernels
    """
    def __init__(self, feature_maps: List[FeatureMap]):
        """
        :param feature_maps: Feature maps whose kernels should be multiplied.
        """
        # product features are too large, don't precompute in feature space
        feature_counts = [fm.get_n_features() for fm in feature_maps]
        has_finitely_many_features = all([n >= 0 for n in feature_counts])
        # allow precomp if all but one feature maps have only 1 feature, none has infinitely many,
        # and all can be precomputed
        # this can occur for the gradient feature map in the last layer of a NN
        allow_precompute_features = has_finitely_many_features \
                                    and (len([fc for fc in feature_counts if fc != 1]) <= 1) \
                                    and all([fm.allow_precompute_features for fm in feature_maps])
        super().__init__(n_features=utils.prod(feature_counts) if has_finitely_many_features else -1,
                         allow_precompute_features=allow_precompute_features)
        self.feature_maps = feature_maps

    def precompute_soft_(self, feature_data: FeatureData, idxs: Indexes) -> Tuple['FeatureMap', FeatureData]:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        results = [fm.precompute(fd, idxs) for fm, fd in zip(self.feature_maps, feature_data.feature_data_list)]
        return ProductFeatureMap([r[0] for r in results]), ListFeatureData([r[1] for r in results])

    def get_feature_matrix_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        # raise RuntimeError('Computing feature matrix for product features!')
        slices = [[slice(None)] + [None] * i + [slice(None)] + [None] * (len(self.feature_maps) - 1 - i)
                  for i in range(len(self.feature_maps))]
        matrix_long = utils.prod([fm.get_feature_matrix(fd, idxs)[s]
                                  for fm, fd, s in zip(self.feature_maps, feature_data.feature_data_list, slices)])
        return matrix_long.reshape(matrix_long.shape[0], -1)

    def get_kernel_matrix_impl_(self, feature_data_1: FeatureData, feature_data_2: FeatureData,
                                idxs_1: Indexes, idxs_2: Indexes) -> torch.Tensor:
        if not isinstance(feature_data_1, ListFeatureData):
            raise ValueError(f'feature_data_1 must be of type ListFeatureData, but is of type {type(feature_data_1)}')
        if not isinstance(feature_data_2, ListFeatureData):
            raise ValueError(f'feature_data_2 must be of type ListFeatureData, but is of type {type(feature_data_2)}')
        return utils.prod([fm.get_kernel_matrix(fd1, fd2, idxs_1, idxs_2)
                           for fm, fd1, fd2 in zip(self.feature_maps, feature_data_1.feature_data_list,
                                                   feature_data_2.feature_data_list)])

    def get_kernel_matrix_diag_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        return utils.prod([fm.get_kernel_matrix_diag(fd, idxs)
                           for fm, fd in zip(self.feature_maps, feature_data.feature_data_list)])

    def sketch(self, n_features: int, **config) -> 'FeatureMap':
        # use a simple tensor sketch, more complicated sketches could be used
        return ElementwiseProductFeatureMap([fm.sketch(n_features, **config) for fm in self.feature_maps])


class ElementwiseSumFeatureMap(FeatureMap):
    """
    Feature map representing the sum of multiple feature maps with equal feature space dimension.
    This is used for sketching SumFeatureMaps.
    """
    def __init__(self, feature_maps: List[FeatureMap]):
        """
        :param feature_maps: Feature maps whose features should be added.
        """
        super().__init__(n_features=feature_maps[0].get_n_features(),
                         allow_precompute_features=all([fm.allow_precompute_features for fm in feature_maps]))
        self.feature_maps = feature_maps

    def precompute_soft_(self, feature_data: FeatureData, idxs: Indexes) -> Tuple['FeatureMap', FeatureData]:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        results = [fm.precompute(fd, idxs) for fm, fd in zip(self.feature_maps, feature_data.feature_data_list)]
        return ElementwiseSumFeatureMap([r[0] for r in results]), ListFeatureData([r[1] for r in results])

    def get_feature_matrix_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        return sum(
            [fm.get_feature_matrix(fd, idxs) for fm, fd in zip(self.feature_maps, feature_data.feature_data_list)])


class ElementwiseProductFeatureMap(FeatureMap):
    """
    Feature map representing the elementwise product of multiple feature maps with equal feature space dimension.
    This is used for sketching ProductFeatureMaps.
    """
    def __init__(self, feature_maps: List[FeatureMap]):
        """
        :param feature_maps: Feature maps that should be multiplied element-wise.
        """
        super().__init__(n_features=feature_maps[0].get_n_features(),
                         allow_precompute_features=all([fm.allow_precompute_features for fm in feature_maps]))
        self.feature_maps = feature_maps

    def precompute_soft_(self, feature_data: FeatureData, idxs: Indexes) -> Tuple['FeatureMap', FeatureData]:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        results = [fm.precompute(fd, idxs) for fm, fd in zip(self.feature_maps, feature_data.feature_data_list)]
        return ElementwiseProductFeatureMap([r[0] for r in results]), ListFeatureData([r[1] for r in results])

    def get_feature_matrix_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        if not isinstance(feature_data, ListFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        return self.n_features ** ((len(self.feature_maps) - 1) / 2) \
               * utils.prod([fm.get_feature_matrix(fd, idxs)
                             for fm, fd in zip(self.feature_maps, feature_data.feature_data_list)])


class LinearFeatureMap(FeatureMap):
    """
    Feature map of the form phi(x) = Ax for a matrix A.
    """
    def __init__(self, matrix: torch.Tensor):
        """
        :param matrix: should be in_features x out_features, i.e. transposed
        """
        super().__init__(n_features=matrix.shape[-1])
        self.matrix = matrix

    def get_feature_matrix_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        if not isinstance(feature_data, TensorFeatureData):
            raise ValueError(f'feature_data must be of type ListFeatureData, but is of type {type(feature_data)}')
        feature_matrix = feature_data.get_tensor(idxs)
        self.matrix = self.matrix.to(feature_matrix.device)
        self.matrix = self.matrix.type(feature_matrix.type())
        return feature_matrix.matmul(self.matrix)

    def sketch(self, n_features: int, **config) -> 'FeatureMap':
        # sketch by using the sketching function of IdentityFeatureMap
        return SequentialFeatureMap(
            IdentityFeatureMap(n_features=self.get_n_features()), [self]).sketch(n_features, **config)


class SequentialFeatureMap(FeatureMap):
    """
    Represents a feature map of the form phi(x) = f(g(x)) or even more concatenated functions.
    Here, g does not need to be a feature map but only a DataTransform.
    """
    def __init__(self, feature_map: FeatureMap, tfms: List[Callable]):
        """
        :param feature_map: Feature map that is applied last, after all the transforms.
        :param tfms: Transforms, in the order that they should be applied to the input.
        """
        super().__init__(n_features=feature_map.get_n_features(),
                         allow_precompute_features=feature_map.allow_precompute_features)
        self.feature_map = feature_map
        self.tfms = tfms

    def precompute_soft_(self, feature_data: FeatureData, idxs: Indexes) -> Tuple['FeatureMap', FeatureData]:
        feature_data = SubsetFeatureData(feature_data, idxs)
        for tfm in self.tfms:
            feature_data = tfm(feature_data)
        return self.feature_map.precompute(feature_data)

    def get_feature_matrix_impl_(self, feature_data: FeatureData, idxs: Indexes) -> torch.Tensor:
        feature_data = self.tfms[0](feature_data, idxs)  # or: SubsetFeatureData(feature_data, idxs)
        for tfm in self.tfms[1:]:
            feature_data = tfm(feature_data)
        return self.feature_map.get_feature_matrix(feature_data)

    def sketch(self, n_features: int, **config) -> 'FeatureMap':
        return SequentialFeatureMap(self.feature_map.sketch(n_features, **config), self.tfms)


class ToDoubleTransform(DataTransform):
    """
    Transforms data to float64 format
    """
    def forward(self, feature_data: FeatureData, idxs: Indexes) -> FeatureData:
        return SubsetFeatureData(feature_data.cast_to(torch.float64), idxs)



# This file allows to configure where to save data, results, plots etc.
class CustomPaths:
    # path where downloaded data sets will be saved
    data_path = '/media/data/prog/projects/batch_active_learning'
    # path where benchmark results will be saved
    results_path = '/home/david/Dokumente/batch_active_learning/results'
    # path where plots and tables will be saved
    plots_path = 'plots'
    # path where benchmark results can be cached in a more efficient format such that they load faster
    cache_path = 'cache'


def get_data_path():
    return CustomPaths.data_path


def get_results_path():
    return CustomPaths.results_path


def get_plots_path():
    return CustomPaths.plots_path


def get_cache_path():
    return CustomPaths.cache_path


import os
import os.path
import heapq
import glob
import gzip
import shutil
import copy
import timeit
import time
import numpy as np
from typing import *
import dill
import json
import itertools


def select_from_config(config, keys):
    selected = {}
    for key in keys:
        if key in config:
            selected[key] = config[key]
    return selected


def adapt_config(config, **kwargs):
    new_config = copy.deepcopy(config)
    for key, value in kwargs.items():
        new_config[key] = value
    return new_config


def existsDir(directory):
    if directory != '':
        if not os.path.exists(directory):
            return False
    return True


def existsFile(file_path):
    return os.path.isfile(file_path)


def ensureDir(file_path):
    directory = os.path.dirname(file_path)
    if directory != '':
        if not os.path.exists(directory):
            os.makedirs(directory)


def matchFiles(file_matcher):
    return glob.glob(file_matcher)


def newDirname(prefix):
    i = 0
    name = prefix
    if existsDir(prefix):
        while existsDir(prefix + "_" + str(i)):
            i += 1
        name = prefix + "_" + str(i)
    os.makedirs(name)
    return name


def getSubfolderNames(folder):
    return [os.path.basename(name)
            for name in os.listdir(folder)
            if os.path.isdir(os.path.join(folder, name))]


def getSubfolders(folder):
    return [os.path.join(folder, name)
            for name in os.listdir(folder)
            if os.path.isdir(os.path.join(folder, name))]


def writeToFile(filename, content):
    ensureDir(filename)
    file = open(filename, 'w')
    file.truncate()
    file.write(content)
    file.close()


def readFromFile(filename):
    if not os.path.isfile(filename):
        return ''

    file = open(filename, 'r')
    result = file.read()
    file.close()
    return result


def serialize(filename, obj, compressed=False, use_json=False):
    # json only works for nested dicts
    ensureDir(filename)
    if compressed:
        file = gzip.open(filename, 'w' if use_json else 'wb')
    else:
        file = open(filename, 'w' if use_json else 'wb')
    # dill can dump lambdas, and dill also dumps the class and not only the contents
    if use_json:
        json.dump(obj, file)
    else:
        dill.dump(obj, file)
    file.close()


def deserialize(filename, compressed=False, use_json=False):
    # json only works for nested dicts
    if compressed:
        file = gzip.open(filename, 'r' if use_json else 'rb')
    else:
        file = open(filename, 'r' if use_json else 'rb')
    if use_json:
        result = json.load(file)
    else:
        result = dill.load(file)
    file.close()
    return result


def copyFile(src, dst):
    ensureDir(dst)
    shutil.copyfile(src, dst)


def nsmallest(n, inputList):
    return heapq.nsmallest(n, inputList)[-1]


def identity(x):
    return x


def set_none_except(lst, idxs):
    for i in range(len(lst)):
        if i not in idxs:
            lst[i] = None


def argsort(lst, key: Optional[Callable] = None):
    # from https://stackoverflow.com/questions/3382352/equivalent-of-numpy-argsort-in-basic-python
    if key is None:
        range_key = lst.__getitem__
    else:
        range_key = lambda i, f=key: f(lst[i])
    return sorted(range(len(lst)), key=range_key)


def dict_argsort(dict_to_sort: dict, key: Optional[Callable] = None):
    keys = list(dict_to_sort.keys())
    values = list(dict_to_sort.values())
    perm = argsort(values, key=key)
    return [keys[i] for i in perm]


def join_dicts(*dicts):
    # Attention: arguments do not commute since later dicts can override entries from earlier dicts!
    result = copy.copy(dicts[0])
    for d in dicts[1:]:
        result.update(d)
    return result


def update_dict(d: dict, update: Optional[dict] = None, remove_keys: Optional[Union[object, List[object]]] = None):
    d = copy.copy(d)
    if update is not None:
        d.update(update)
    if remove_keys is not None:
        if isinstance(remove_keys, List):
            for key in remove_keys:
                d.pop(key)
        else:
            d.pop(remove_keys)
    return d


def pretty_table_str(str_table):
    max_lens = [np.max([len(row[i]) for row in str_table])for i in range(len(str_table[0]))]
    whole_str = ''
    for row in str_table:
        for i, entry in enumerate(row):
            whole_str += entry + (' ' * (max_lens[i] - len(entry)))
        whole_str += '\n'
    return whole_str[:-1]  # remove last newline


def prod(it: Iterable, id=None):
    result = None
    for value in it:
        if result is None:
            result = value
        else:
            result = result * value
    if result is None:
        if id is None:
            raise ValueError(f'Cannot compute empty product without identity element')
        else:
            return id
    return result


def all_equal(it: Iterable):
    # see https://stackoverflow.com/questions/3844801/check-if-all-elements-in-a-list-are-identical
    g = itertools.groupby(it)  # iterates over unique elements
    try:
        next(g)
        next(g)
    except StopIteration:
        return True
    return False


class Timer:
    def __init__(self):
        self.start_time_total = None
        self.start_time_process = None
        self.acc_time_total = 0.0
        self.acc_time_process = 0.0

    def start(self):
        if self.start_time_total is None or self.start_time_process is None:
            self.start_time_total = timeit.default_timer()
            self.start_time_process = time.process_time()

    def pause(self):
        if self.start_time_total is None or self.start_time_process is None:
            return  # has already been paused or not been started
        self.acc_time_total += timeit.default_timer() - self.start_time_total
        self.acc_time_process += time.process_time() - self.start_time_process
        self.start_time_total = None
        self.start_time_process = None

    def get_result_dict(self):
        return {'total': self.acc_time_total, 'process': self.acc_time_process}


class TimePrinter:
    def __init__(self, desc: str):
        self.desc = desc
        self.timer = Timer()

    def __enter__(self):
        self.timer.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.timer.pause()
        print(f'Time for {self.desc}: {self.timer.get_result_dict()["total"]:g}s')


def format_length_s(duration: float) -> str:
    seconds = int(duration)
    minutes = seconds // 60
    seconds -= minutes * 60
    hours = minutes // 60
    minutes -= hours * 60
    days = hours // 24
    hours -= days * 24

    result = f'{seconds}s'
    if minutes > 0:
        result = f'{minutes}m' + result
    if hours > 0:
        result = f'{hours}h' + result
    if days > 0:
        result = f'{days}d' + result

    return result


def format_date_s(time_s: float) -> str:
    return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time_s))


def get_batch_intervals(n_total: int, batch_size: int) -> List[Tuple[int, int]]:
    boundaries = [i * batch_size for i in range(1 + n_total // batch_size)]
    if boundaries[-1] != n_total:
        boundaries.append(n_total)
    return [(start, stop) for start, stop in zip(boundaries[:-1], boundaries[1:])]


def last_mod_time_recursive(path: str):
    # see https://stackoverflow.com/questions/29685069/get-the-last-modified-date-of-a-directory-including-subdirectories-using-pytho
    import os
    return max(os.path.getmtime(root) for root, _, _ in os.walk(path))








from typing import *
import torch

from .. import utils


def torch_cat(tensors: List[torch.Tensor], dim: int):
    """
    Implements torch.cat() but doesn't copy if only one tensor is provided.
    This can make it faster if no copying behavior is needed.
    :param tensors: Tensors to be concatenated.
    :param dim: Dimension in which the tensor should be concatenated.
    :return: The concatendated tensor.
    """
    if len(tensors) == 1:
        return tensors[0]
    return torch.cat(tensors, dim=dim)


class Indexes:
    """
    This class allows to store and manipulate ways of indexing tensors, either via slices or lists of integers.
    This is important for computing subsets of kernel matrices and minibatching of such computations.
    """
    def __init__(self, n_samples: int, idxs: Optional[Union[torch.Tensor, slice, int, 'Indexes']]):
        """
        :param n_samples: Total size of the dimension of tensors that should be indexed.
        This is used to convert negative slices like [:-1] to positive slices like [:n_samples-1].
        Also, it is used to compute how many elements indexing with the Indexes object will be returned.
        :param idxs: Object that a tensor should be indexed with. Can be either
        - a tensor of ints,
        - a non-empty slice without stride, i.e., [1:2:-1] is not allowed because of the -1,
                and [2:1] is not allowed because it is empty.
        - a single index i (which will be treated as the slice [i:i+1],
                 i.e., it will not remove the dimension from the indexed tensor),
        - None, corresponding to [:], i.e., selecting all elements
        - or already an Indexes object (which will then be copied)
        """
        self.n_samples = n_samples
        if idxs is None:
            self.idxs = slice(0, n_samples)
        elif isinstance(idxs, Indexes):
            self.idxs = idxs.idxs
        elif isinstance(idxs, int):
            self.idxs = slice(idxs, idxs+1)
        elif isinstance(idxs, slice):
            if idxs.step is not None and idxs.step != 1:
                raise ValueError(f'Cannot handle slices with step size other than 1')
            start = 0 if idxs.start is None else idxs.start + (0 if idxs.start >= 0 else n_samples)
            stop = n_samples if idxs.stop is None else idxs.stop + (0 if idxs.stop >= 0 else n_samples)
            if stop <= start:
                raise ValueError(f'stop <= start not allowed for slices')
            self.idxs = slice(start, stop)
        elif isinstance(idxs, torch.Tensor):
            self.idxs = idxs
        else:
            raise ValueError(f'Cannot handle index type {type(idxs)}')

        if isinstance(self.idxs, slice):
            self.n_idxs = self.idxs.stop - self.idxs.start
        else:
            self.n_idxs = len(self.idxs)

    def __len__(self):
        """
        :return: Number of elements that remain after indexing with self
        """
        return self.n_idxs

    def compose(self, other: 'Indexes') -> 'Indexes':
        """
        :param other: an Indexes object that should be used to subselect indexes from self
        :return: Returns an Indexes object such that data[self][other] = data[self.compose(other)]
        """
        if isinstance(self.idxs, torch.Tensor):
            new_idxs = self.idxs[other.idxs]
        elif isinstance(self.idxs, slice):
            if isinstance(other.idxs, torch.Tensor):
                new_idxs = other.idxs + self.idxs.start
            elif isinstance(other.idxs, slice):
                new_idxs = slice(self.idxs.start + other.idxs.start, self.idxs.start + other.idxs.stop)
            else:
                raise RuntimeError('other.idxs is neither slice nor torch.Tensor')
        else:
            raise RuntimeError('self.idxs is neither slice nor torch.Tensor')

        return Indexes(len(self), new_idxs)

    def get_idxs(self):
        """
        :return: Returns the slice or LongTensor that this object represents, which can be used to index a torch.Tensor
        """
        return self.idxs

    def split_by_sizes(self, sample_sizes: List[int]) -> Iterable[Tuple[int, 'Indexes']]:
        """
        This method allows to apply the Indexes object to multiple arrays as if they were concatenated.
        This is useful for indexing batched data.
        This method is currently only implemented for Indexes that represent slices
        and not those that represent LongTensors.
        Example:
        full_arr = torch.arange(32)
        idxs = Indexes(16, slice(5, 17))
        batched_arrs = [full_arr[4*i:4*(i+1)] for i in range(8)]
        batch_sizes = [len(arr) for arr in batched_arrs]
        full_arr_indexed = full_arr[idxs.get_idxs()]
        batched_arrs_indexed = torch.concat([batched_arrs[i][sub_idxs.get_idxs()]
                                        for i, sub_idxs in idxs.split_by_sizes(batch_sizes)], dim=0)
        assert full_arr_indexed == batched_arrs_indexed
        :param sample_sizes: Sizes of the multiple arrays that should be indexed
        :return: Returns a generator object that can be used to iterate over sub-Indexes objects
        corresponding to the corresponding data parts.
        """
        if isinstance(self.idxs, slice):
            start = self.idxs.start
            stop = self.idxs.stop
            for i, sz in enumerate(sample_sizes):
                if start < sz and stop > 0:
                    yield i, Indexes(sz, slice(max(start, 0), min(stop, sz)))
                start -= sz
                stop -= sz
        else:
            raise NotImplementedError('indexing splitted data with a list of indexes is currently not supported')

    def is_all_slice(self) -> bool:
        """
        :return: Returns true iff applying this object is equivalent to selecting all data, i.e. [:].
        """
        return isinstance(self.idxs, slice) and self.idxs.start == 0 and self.idxs.stop == self.n_samples


class FeatureData:
    """
    Abstract base class for classes that represent data that serves as input to feature maps.
    """
    def __init__(self, n_samples: int, device: str, dtype):
        """
        :param n_samples: Number of samples represented by this FeatureData object.
        :param device: String representing the torch device that this feature data is located on.
        :param dtype: Torch dtype that the feature data has
        """
        self.n_samples = n_samples
        self.device = device
        self.dtype = dtype

    def get_n_samples(self) -> int:
        """
        :return: Returns the number of samples.
        """
        return self.n_samples

    def get_device(self) -> str:
        """
        :return: Returns the device that the data is on.
        """
        return self.device

    def get_dtype(self) -> Any:
        """
        :return: Returns the (torch) dtype that the feature data has.
        """
        return self.dtype

    def __len__(self) -> int:
        """
        :return: Returns the number of samples.
        """
        return self.get_n_samples()

    def __getitem__(self, idxs: Union[torch.Tensor, slice, int, 'Indexes']):
        """
        :param idxs: Represents the subset of samples that should be returned.
        Note that if idxs is an int, the dimension will not be collapsed.
        In other words, self[i] is equivalent to self[i:i+1].
        :return: Returns the feature data represented by the subset of indexes in idxs.
        """
        idxs = Indexes(self.get_n_samples(), idxs)
        if idxs.is_all_slice():
            return self
        return SubsetFeatureData(self, idxs)

    def simplify(self, idxs: Optional[Union[torch.Tensor, slice, int, Indexes]] = None) -> 'FeatureData':
        """
        Simplifies the FeatureData object recursively, for example by concatenating batches.
        Do not override this method in subclasses. Override simplify_impl_() instead.
        :param idxs: The subset of the data that should be simplified.
        :return: Returns a simplified version of self[idxs].
        """
        idxs = Indexes(self.get_n_samples(), idxs)
        return self.simplify_impl_(idxs)

    def simplify_impl_(self, idxs: Indexes) -> 'FeatureData':
        """
        Internal method to implement simplify(). This method should be overridden by subclasses.
        :param idxs: Indexes for the subset of the data that should be simplified.
        :return: Returns a simplified version of self[idxs].
        """
        raise NotImplementedError()

    def simplify_multi_(self, feature_data_list: List['FeatureData'], idxs_list: List[Indexes]) -> 'FeatureData':
        """
        Internal method that is sometimes required to implement simplify() recursively.
        This method should be overridden by subclasses.
        This method should simplify feature_data[idxs] for feature_data, idxs in zip(feature_data_list, idxs_list)
        and then concatenate the feature_data objects suitably.
        The concatenation needs to be handled by this custom method
        because we want the concatenation of multiple ListFeatureData objects
        to be a ListFeatureData where the corresponding list elements get concatenated.
        :param feature_data_list: Lists of feature data objects to be simplified and concatenated.
        The feature data objects should be of the same subclass of FeatureData as self.
        :param idxs_list: List of indexes to apply to the feature data objects.
        :return: A simplified version of a concatenation of the indexed feature_data_list elements.
        """
        raise NotImplementedError()

    def iterate(self, idxs: Indexes) -> Iterable[Tuple[Indexes, 'FeatureData']]:
        """
        This method allows to iterate over sub-parts of the feature data.
        This is necessary because the FeatureData may consist of a concatenation
        of multiple individual tensor feature data objects.
        Example:
        tfd_list = [TensorFeatureData(torch.zeros(5, 3)) for i in range(3)]
        feature_data = ConcatFeatureData(tfd_list)
        idxs = Indexes(len(feature_data), None)
        tensor = torch.concat([tfd.get_tensor(sub_idxs) for sub_idxs, tfd in feature_data.iterate(idxs)], dim=0)
        # in this case, one could have just used tensor = feature_data.get_tensor(idxs) directly
        # however, this is useful for computing kernel matrices in batches, for example.
        :param idxs: Sub-indexes of the data that should be iterated over.
        :return: Returns a generator object that allows to iterate over tuples of (Indexes, FeatureData).
        """
        yield idxs, self  # default implementation

    def __iter__(self) -> Iterable[Tuple[Indexes, 'FeatureData']]:
        """
        This method corresponds to iterate(Indexes(len(self), None)), i.e., iterating over the whole data.
        To use this method, just use something like "for sub_idxs, sub_data in feature_data", without the .__iter__().
        :return: Returns a generator object that allows to iterate over parts of the data, see iterate().
        """
        return self.iterate(Indexes(n_samples=self.get_n_samples(), idxs=None))

    def batched(self, batch_size: int) -> 'FeatureData':
        """
        Batch this feature data, such that __iter__() or iterate() iterates over this data in multiple batches
        :param batch_size: Maximum batch size (the last batch may be smaller).
        :return: A BatchedFeatureData object.
        """
        return BatchedFeatureData(self, batch_size=batch_size)

    def get_tensor(self, idxs: Optional[Union[torch.Tensor, slice, int, Indexes]] = None) -> torch.Tensor:
        """
        Returns the tensor corresponding to this feature data indexed by idxs.
        This method is only well-defined for feature data that does not involve ListFeatureData,
        since otherwise it would need to return a list of tensors.
        This method should not be overridden by subclasses, they should override get_tensor_impl_() instead.
        :param idxs: object to index the tensor with, or None if the full tensor should be returned.
        :return: Returns a torch.Tensor representing the indexed data.
        The first dimension of the tensor is the n_samples dimension.
        """
        idxs = Indexes(self.get_n_samples(), idxs)
        return self.get_tensor_impl_(idxs)

    def get_tensor_impl_(self, idxs: Indexes) -> torch.Tensor:
        """
        Internal method for computing the result of get_tensor().
        Should be overridden by classes
        that do not want to use the default behavior that is implemented here via self.iterate(idxs)
        :param idxs: Indexes object to index the tensor with.
        :return: The resulting tensor.
        """
        return torch_cat([feature_data.get_tensor(sub_idxs) for sub_idxs, feature_data in self.iterate(idxs)], dim=0)

    def cast_to(self, dtype) -> 'FeatureData':
        """
        Casts the data to another dtype.
        :param dtype: Target type, should be usable by torch.
        :return: A FeatureData object with internal data cast to the target dtype.
        """
        raise NotImplementedError()

    def to_indexes(self, idxs: Optional[Indexes]) -> Indexes:
        """
        :param idxs: Indexes object or None
        :return: the Indexes object or an Indexes object representing None
        """
        return idxs if idxs is not None else Indexes(self.get_n_samples(), None)


class EmptyFeatureData(FeatureData):
    """
    This object represents feature data with no samples.
    This is used internally during simplification and should not matter to users.
    Some methods like get_tensor() may not work for this class.
    """
    def __init__(self, device: str, dtype):
        super().__init__(n_samples=0, device=device, dtype=dtype)

    def simplify_impl_(self, idxs: Indexes) -> 'FeatureData':
        return self

    def simplify_multi_(self, feature_data_list: List['EmptyFeatureData'], idxs_list: List[Indexes]) -> 'FeatureData':
        return EmptyFeatureData(device=self.device, dtype=self.dtype)

    def cast_to(self, dtype) -> 'FeatureData':
        return self


class TensorFeatureData(FeatureData):
    """
    FeatureData subclass representing data consisting of a single tensor of shape [n_samples, n_features].
    """
    def __init__(self, data: torch.Tensor):
        """
        :param data: Tensor of shape [n_samples, ...], usually [n_samples, n_features]
        """
        super().__init__(n_samples=data.shape[-2], device=data.device, dtype=data.dtype)
        self.data = data

    def get_tensor_impl_(self, idxs: Indexes) -> torch.Tensor:
        return self.data[idxs.get_idxs()]

    def simplify_impl_(self, idxs: Indexes) -> 'FeatureData':
        return TensorFeatureData(self.get_tensor_impl_(idxs))

    def simplify_multi_(self, feature_data_list: List['TensorFeatureData'], idxs_list: List[Indexes]) -> 'FeatureData':
        return TensorFeatureData(torch_cat([fd.data[idxs.get_idxs()]
                                            for fd, idxs in zip(feature_data_list, idxs_list)], dim=0))

    def cast_to(self, dtype) -> 'FeatureData':
        return TensorFeatureData(self.data.type(dtype))


class SubsetFeatureData(FeatureData):
    """
    FeatureData subclass representing a subset of other FeatureData,
    where the subset is given by an Indexes object.
    """
    def __init__(self, feature_data: FeatureData, idxs: Indexes):
        """
        :param feature_data: FeatureData which this object represents a subset of
        :param idxs: Indexes object representing the subset of indexes.
        """
        super().__init__(n_samples=len(idxs), device=feature_data.get_device(), dtype=feature_data.get_dtype())
        self.idxs = idxs
        self.feature_data = feature_data

    def iterate(self, idxs: Optional[Indexes] = None) -> Iterable[Tuple[Indexes, 'FeatureData']]:
        for sub_idxs, feature_data in self.feature_data.iterate(self.idxs.compose(idxs)):
            yield sub_idxs, feature_data

    def simplify_impl_(self, idxs: Indexes) -> 'FeatureData':
        return self.feature_data.simplify(self.idxs.compose(idxs))

    def simplify_multi_(self, feature_data_list: List['SubsetFeatureData'], idxs_list: List[Indexes]) -> 'FeatureData':
        return ConcatFeatureData([fd.simplify(idxs) for fd, idxs in zip(feature_data_list, idxs_list)]).simplify()

    def cast_to(self, dtype) -> 'FeatureData':
        return SubsetFeatureData(self.feature_data.cast_to(dtype), self.idxs)


class ConcatFeatureData(FeatureData):
    """
    FeatureData subclass that allows to concatenate multiple FeatureData objects into one object.
    Note that this does not concatenate the tensors,
    and .iterate() / .__iter__() may therefore loop over multiple feature data objects.
    While this class can be used to force batched processing of data,
    it is recommended to use BatchedFeatureData for this purpose since it does not actually split up the tensors.
    """
    def __init__(self, feature_data_list: List[FeatureData]):
        """
        :param feature_data_list: List of feature data objects that should be (virtually) concatenated.
        """
        sample_sizes = [fd.get_n_samples() for fd in feature_data_list]
        super().__init__(n_samples=sum(sample_sizes), device=feature_data_list[0].get_device(),
                         dtype=feature_data_list[0].get_dtype())
        self.feature_data_list = feature_data_list
        self.sample_sizes = sample_sizes

    def iterate(self, idxs: Indexes) -> Iterable[Tuple[Indexes, 'FeatureData']]:
        for i, sub_idxs in idxs.split_by_sizes(self.sample_sizes):
            for sub_idxs_2, feature_data in self.feature_data_list[i].iterate(sub_idxs):
                yield sub_idxs_2, feature_data

    def simplify_impl_(self, idxs: Indexes) -> 'FeatureData':
        simplified = [self.feature_data_list[i].simplify(sub_idxs)
                      for i, sub_idxs in idxs.split_by_sizes(self.sample_sizes)]
        simplified = [fd for fd in simplified if not isinstance(fd, EmptyFeatureData)]
        if len(simplified) == 1:
            return simplified[0]
        elif len(simplified) == 0:
            return EmptyFeatureData(device=self.device)
        else:
            if utils.all_equal([type(fd) for fd in simplified]):
                # call simplify_multi_ of the sub-data
                # for example, if simplified[0] is a ListFeatureData object,
                # it will need to concatenate list elements separately
                return simplified[0].simplify_multi_(simplified, [Indexes(fd.get_n_samples(), None) for fd in simplified])
            else:
                # this should not occur in practice
                # because all concatenated FeatureData objects should have the same basic type
                raise RuntimeError(
                    'Attempting to concatenate different-typed simplified FeatureData objects during simplify')

    def simplify_multi_(self, feature_data_list: List['ConcatFeatureData'], idxs_list: List[Indexes]) -> 'FeatureData':
        if len(feature_data_list) == 0:
            return EmptyFeatureData(device=self.device)
        # use simplify() to implement simplify_multi_()
        return ConcatFeatureData([SubsetFeatureData(fd, idxs) for cfd, idxs in zip(feature_data_list, idxs_list)
                                  for fd in cfd.feature_data_list]).simplify()

    def cast_to(self, dtype) -> 'FeatureData':
        return ConcatFeatureData([fd.cast_to(dtype) for fd in self.feature_data_list])


class BatchedFeatureData(FeatureData):
    """
    This class can be used to force batched processing of data,
    by iterating over batches in its .iterate() and .__iter__() methods.
    In contrast to ConcatFeatureData, it does only pretend that it contains multiple sub FeatureData objects.
    Since it only holds one FeatureData object, it is more efficient to use for simplify() and get_tensor().
    """
    def __init__(self, feature_data: FeatureData, batch_size: int):
        """
        :param feature_data: FeatureData to be batched.
        :param batch_size: Batch size. The last batch is smaller
        if len(feature_data) is not divisible by batch_size.
        """
        super().__init__(n_samples=len(feature_data), device=feature_data.get_device(), dtype=feature_data.get_dtype())
        self.feature_data = feature_data
        self.batch_size = batch_size
        batch_intervals = utils.get_batch_intervals(self.n_samples, batch_size=batch_size)
        self.batch_idxs = [Indexes(self.n_samples, slice(start, stop)) for start, stop in batch_intervals]
        self.sample_sizes = [len(idxs) for idxs in self.batch_idxs]

    def iterate(self, idxs: Indexes) -> Iterable[Tuple[Indexes, 'FeatureData']]:
        # print(f'{idxs.get_idxs()=}, {list(idxs.split_by_sizes(self.sample_sizes))=}, {self.sample_sizes=}')
        # print(f'{idxs.get_idxs()=}, {self.sample_sizes=}')
        for i, sub_idxs in idxs.split_by_sizes(self.sample_sizes):
            for sub_idxs_2, feature_data in self.feature_data.iterate(self.batch_idxs[i].compose(sub_idxs)):
                yield sub_idxs_2, feature_data

    def simplify_impl_(self, idxs: Indexes) -> 'FeatureData':
        return self.feature_data.simplify(idxs)

    def simplify_multi_(self, feature_data_list: List['BatchedFeatureData'], idxs_list: List[Indexes]) -> 'FeatureData':
        if len(feature_data_list) == 0:
            return EmptyFeatureData(device=self.device)
        fd_list = [fd.feature_data for fd in feature_data_list]
        return fd_list[0].simplify_multi_(fd_list, idxs_list)

    def cast_to(self, dtype) -> 'FeatureData':
        return BatchedFeatureData(self.feature_data.cast_to(dtype), self.batch_size)


class ListFeatureData(FeatureData):  # does not concatenate along batch dimension
    """
    This class represents a list of separate FeatureData objects.
    All FeatureData objects must have the same number of samples.
    This is different to ConcatFeatureData,
    which assumes that the FeatureData objects are concatenated along the samples dimension.
    """
    def __init__(self, feature_data_list: List[FeatureData]):
        """
        :param feature_data_list: List of feature data objects.
        """
        super().__init__(n_samples=feature_data_list[0].get_n_samples(), device=feature_data_list[0].get_device(),
                         dtype=feature_data_list[0].get_dtype())
        self.feature_data_list = feature_data_list

    def get_tensor(self, idxs: Optional[Union[torch.Tensor, slice, int, Indexes]] = None) -> torch.Tensor:
        raise NotImplementedError(
            'get_tensor() cannot be called on ListFeatureData since it would need to return multiple tensors')

    def simplify_impl_(self, idxs: Indexes) -> 'FeatureData':
        return ListFeatureData([fd.simplify(idxs) for fd in self.feature_data_list])

    def simplify_multi_(self, feature_data_list: List['ListFeatureData'], idxs_list: List[Indexes]) -> 'FeatureData':
        if len(feature_data_list) == 0:
            return EmptyFeatureData(device=self.device, dtype=self.dtype)
        return ListFeatureData([ConcatFeatureData([fd.feature_data_list[i][idxs]
                                                   for fd, idxs in zip(feature_data_list, idxs_list)]).simplify()
                                for i in range(len(feature_data_list[0].feature_data_list))])

    def cast_to(self, dtype) -> 'FeatureData':
        return ListFeatureData([fd.cast_to(dtype) for fd in self.feature_data_list])



from multiprocessing.pool import Pool
import multiprocessing as mp

from .train import *
from .data import *
from .task_execution import get_devices

# This file can be used to check how much the error for random selection drops on different data sets
# when increasing the training set size.
# We used this to remove data sets where increasing the number of samples does not have a strong effect.


def single_learnability(args):
    task, task_id, device_idx = args
    # task = Task.get_tabular_tasks(n_train=10, al_batch_sizes=[2**k for k in range(9)] + [256]*10)[3]
    trainer = ModelTrainer('NN_test', save=False)
    return trainer(TaskSplit(task, id=task_id), device=get_devices()[device_idx])


def check_task_learnability(n_repeats=5, n_train_small=1000, n_train_large=5000, task_names=None,
                            n_processes_per_device=5):
    small_tasks = Task.get_tabular_tasks(n_train=n_train_small, al_batch_sizes=[], ds_names=task_names)
    large_tasks = Task.get_tabular_tasks(n_train=n_train_large, al_batch_sizes=[], ds_names=task_names)
    n_devices = len(get_devices())
    # task = Task.get_tabular_tasks(n_train=10, al_batch_sizes=[2**k for k in range(9)] + [256]*10)[3]
    # trainer = ModelTrainer('NN_test', save=False, batch_size=512)
    for small_task, large_task in zip(small_tasks, large_tasks):
        assert(small_task.task_name == large_task.task_name)
        pool = Pool(processes=min(n_repeats, n_devices * n_processes_per_device))
        small_results = pool.map(single_learnability,
                                 [(small_task, task_id, task_id % n_devices) for task_id in range(n_repeats)],
                                 chunksize=1)
        pool = Pool(processes=min(n_repeats, n_devices * n_processes_per_device))
        large_results = pool.map(single_learnability,
                                 [(large_task, task_id, task_id % n_devices) for task_id in range(n_repeats)],
                                 chunksize=1)
        # small_results = [trainer(TaskSplit(small_task, id=task_id), device=get_devices()[device_idx]) for task_id in range(n_repeats)]
        # large_results = [trainer(TaskSplit(large_task, id=task_id), device=get_devices()[device_idx]) for task_id in range(n_repeats)]
        small_errors = [r['errors']['rmse'][0] for r in small_results]
        large_errors = [r['errors']['rmse'][0] for r in large_results]
        small_mean_rmse = np.mean(small_errors)
        large_mean_rmse = np.mean(large_errors)
        gain_percent = 100 * (small_mean_rmse - large_mean_rmse) / small_mean_rmse
        print(f'Learnability on task {small_task.task_name}: Reduction by {gain_percent:g}% '
              f'from RMSE={np.mean(small_errors):g} +- {np.std(small_errors):g} for n_train={n_train_small} '
              f'to RMSE={np.mean(large_errors):g} +- {np.std(large_errors):g} for n_train={n_train_large}')


if __name__ == '__main__':
    mp.set_start_method('spawn')
    # check_task_learnability()
    # check_task_learnability(n_repeats=20)
    # check_task_learnability(n_repeats=20, task_names=['poker'])
    # check_task_learnability(n_repeats=20, n_train_small=512, n_train_large=512+16*256)
    # check_task_learnability(n_repeats=20, n_train_small=1024, n_train_large=1024+16*256)
    check_task_learnability(n_repeats=20, n_train_small=256, n_train_large=256+16*256)
    pass



from .bmdal.algorithms import BatchSelectorImpl, select_batch
from .models import *
from .data import *
from . import utils
from pathlib import Path


class ModelTrainer:
    def __init__(self, alg_name: str, create_model=create_tabular_model, n_models: int = 1, **config):
        self.alg_name = alg_name
        self.create_model = create_model
        self.config = config
        self.n_models = n_models

    def get_result_file_path(self, exp_name: str, task_name: str, split_id: int) -> Path:
        results_path = Path(custom_paths.get_results_path())
        result_file = results_path / exp_name / task_name / self.alg_name / str(split_id) / 'results.json'
        return result_file

    def __call__(self, task_split: TaskSplit, device: str, do_timing: bool = False):
        # results_path = Path(custom_paths.get_results_path())
        # result_file = results_path / task_split.task_name / self.alg_name / str(task_split.id) / 'results.json'
        al_device = 'cpu' if self.config.get('al_on_cpu', False) else device

        # can lead to imprecise distance calculations otherwise on devices with TF32
        torch.backends.cuda.matmul.allow_tf32 = False

        # if utils.existsFile(result_file) and not self.config.get('rerun', False):
        #     print(f'Results already exist for {self.alg_name} on split {task_split.id} of task {task_split.task_name}',
        #           flush=True)
        #     return

        print(f'Running {self.alg_name} on split {task_split.id} of task {task_split.task_name} on device {device}',
              flush=True)

        np.random.seed(task_split.id)
        torch.manual_seed(task_split.id)

        base_kernel = self.config.get('base_kernel', None)
        # task_split should contain train_idxs, valid_idxs, test_idxs, pool_idxs
        train_idxs = torch.as_tensor(task_split.train_idxs, dtype=torch.int64, device=device)
        valid_idxs = torch.as_tensor(task_split.valid_idxs, dtype=torch.int64, device=device)
        pool_idxs = torch.as_tensor(task_split.pool_idxs, dtype=torch.int64, device=device)
        test_idxs = torch.as_tensor(task_split.test_idxs, dtype=torch.int64, device=device)
        data = task_split.data.to(device)
        n_features = data.tensors['X'].shape[1]

        train_timer = utils.Timer()
        al_timer = utils.Timer()
        al_stats_list = []

        if base_kernel is None:
            # train on random dataset
            model = self.create_model(n_models=self.n_models, n_features=n_features, **self.config).to(device)
            n_train_missing = sum(task_split.al_batch_sizes)
            random_pool_idxs = pool_idxs[torch.randperm(len(task_split.pool_idxs), device=device)[:n_train_missing]]
            train_idxs = torch.cat([train_idxs, random_pool_idxs])
            train_timer.start()
            fit_model(model, data, self.n_models, train_idxs, valid_idxs, **self.config)
            train_timer.pause()
            results = [test_model(model, data, self.n_models, test_idxs)]
        else:
            results = []

            model = self.create_model(n_models=self.n_models, n_features=n_features, **self.config).to(device)
            train_timer.start()
            fit_model(model, data, self.n_models, train_idxs, valid_idxs, **self.config)
            train_timer.pause()
            results.append(test_model(model, data, self.n_models, test_idxs))

            for al_step, al_batch_size in enumerate(task_split.al_batch_sizes):
                print(f'Performing AL step {al_step+1}/{len(task_split.al_batch_sizes)} with n_train={len(train_idxs)}'
                      f', n_pool={len(pool_idxs)}, al_batch_size={al_batch_size}', flush=True)
                single_models = [model.get_single_model(i).to(al_device) for i in range(self.n_models)]
                X = TensorFeatureData(data.tensors['X'].to(al_device))
                feature_data = {'train': X[train_idxs],
                                'pool': X[pool_idxs]}
                y_train = data.tensors['y'][train_idxs].to(al_device)

                al_timer.start()
                new_idxs, al_stats = select_batch(models=single_models, data=feature_data, y_train=y_train,
                                                  use_cuda_synchronize=do_timing,
                                                  **utils.update_dict(self.config, {'batch_size': al_batch_size}))
                al_timer.pause()
                al_stats_list.append(al_stats)
                if str(device).startswith('cuda'):
                    with torch.cuda.device(device):
                        torch.cuda.empty_cache()
                new_idxs = new_idxs.to(device)
                # print(f'{len(new_idxs)=}')
                logical_new_idxs = torch.zeros(pool_idxs.shape[-1], dtype=torch.bool, device=device)
                logical_new_idxs[new_idxs] = True
                train_idxs = torch.cat([train_idxs, pool_idxs[logical_new_idxs]], dim=-1)
                pool_idxs = pool_idxs[~logical_new_idxs]
                # print(f'{train_idxs.shape[-1]=}')
                # print(f'{pool_idxs.shape[-1]=}')

                model = self.create_model(n_models=self.n_models, n_features=n_features, **self.config).to(device)
                train_timer.start()
                fit_model(model, data, self.n_models, train_idxs, valid_idxs, **self.config)
                train_timer.pause()
                results.append(test_model(model, data, self.n_models, test_idxs))

        extended_config = utils.join_dicts(self.config, {'alg_name': self.alg_name, 'n_models': self.n_models})

        results = {'errors': {key: [r[key] for r in results] for key in ['mae', 'rmse', 'maxe', 'q95', 'q99']},
                   'train_times': train_timer.get_result_dict(),
                   'al_times': al_timer.get_result_dict(),
                   'al_stats': al_stats_list,
                   'config': extended_config}

        # if self.config.get('save', True):
        #     utils.serialize(result_file, results, use_json=True)

        print(f'Finished running {self.alg_name} on split {task_split.id} of task {task_split.task_name} on device {device}', flush=True)

        return results


def test_model(model, data, n_models, test_idxs):
    test_dl = ParallelDictDataLoader(data, test_idxs.expand(n_models, -1), batch_size=8192, shuffle=False,
                                     adjust_bs=False, drop_last=False)
    with torch.no_grad():
        model.eval()
        # for batch in test_dl:
        #     print(batch['y'].shape, batch['X'].shape, model(batch['X']).shape)
        errors = torch.cat([torch.abs(batch['y'] - model(batch['X'])) for batch in test_dl], dim=1).squeeze(-1)
    n_models = errors.shape[0]
    mae = errors.mean().item()
    rmse = (errors**2).mean().sqrt().item()
    maxe = torch.max(errors, dim=1)[0].mean().item()
    q95_errors = []
    q99_errors = []
    for i in range(n_models):
        errors_sorted, _ = torch.sort(errors[i])
        q95_errors.append(errors_sorted[int(0.95 * test_idxs.shape[-1])].item())
        q99_errors.append(errors_sorted[int(0.99 * test_idxs.shape[-1])].item())
    q95 = np.mean(q95_errors)
    q99 = np.mean(q99_errors)
    print(f'Test results: MAE={mae:g}, RMSE={rmse:g}, MAXE={maxe:g}, q95={q95:g}, q99={q99:g}')
    print('\n', flush=True)
    return {'mae': mae, 'rmse': rmse, 'maxe': maxe, 'q95': q95, 'q99': q99}


def fit_model(model, data, n_models, train_idxs, valid_idxs, n_epochs=256, batch_size=256, lr=3e-1, weight_decay=0.0,
              valid_batch_size=8192, **config):
    train_dl = ParallelDictDataLoader(data, train_idxs.expand(n_models, -1), batch_size=batch_size, shuffle=True,
                                      adjust_bs=False, drop_last=True)
    valid_dl = ParallelDictDataLoader(data, valid_idxs.expand(n_models, -1), batch_size=valid_batch_size, shuffle=False,
                                      adjust_bs=False, drop_last=False)
    n_steps = n_epochs * len(train_dl)
    best_valid_rmses = [np.Inf] * n_models
    best_model_params = [p.detach().clone() for p in model.parameters()]
    if config.get('opt_name', 'adam') == 'sgd':
        opt = torch.optim.SGD(model.parameters(), lr=lr)
    else:
        if weight_decay > 0.0:
            opt = torch.optim.AdamW(model.parameters(), weight_decay=weight_decay)
        else:
            opt = torch.optim.Adam(model.parameters())
    step = 0
    for i in range(n_epochs):
        # do one training epoch
        # grad_nonzeros = 0
        model.train()
        lr_sched = config.get('lr_sched', 'lin')
        for batch in train_dl:
            X, y = batch['X'], batch['y']
            y_pred = model(X)  # shape: n_models x batch_size x 1
            loss = ((y - y_pred)**2).mean(dim=-1).mean(dim=-1).sum()  # sum over n_models
            loss.backward()
            if lr_sched == 'lin':
                current_lr = lr * (1.0 - step / n_steps)
            elif lr_sched == 'hat':
                current_lr = lr * 2 * (0.5 - np.abs(0.5 - step/n_steps))
            elif lr_sched == 'warmup':
                peak_at = 0.1
                current_lr = lr * min((step/n_steps)/peak_at, (1-step/n_steps)/(1-peak_at))
            else:
                raise ValueError(f'Unknown lr sched "{lr_sched}"')
            for group in opt.param_groups:
                group['lr'] = current_lr
            opt.step()
            with torch.no_grad():
                for param in model.parameters():
                    # grad_nonzeros += torch.count_nonzero(param.grad).item()
                    param.grad = None

            step += 1

        # do one valid epoch
        valid_sses = torch.zeros(n_models, device=data.device)
        model.eval()
        with torch.no_grad():
            # linear_layers = [module for module in model.modules() if isinstance(module, ParallelLinearLayer)]
            # hooks = [ll.register_forward_hook(
            #     lambda layer, inp, out:
            #     print(f'dead neurons: {(out < 0).all(dim=0).all(dim=0).count_nonzero().item()}'))
            #     for ll in linear_layers]
            for batch in valid_dl:
                X, y = batch['X'], batch['y']
                valid_sses = valid_sses + ((y - model(X))**2).mean(dim=-1).sum(dim=-1)
            valid_rmses = torch.sqrt(valid_sses / len(valid_idxs)).detach().cpu().numpy()
            # for hook in hooks:
            #     hook.remove()

        # mean_param_norm = np.mean([p.norm().item() for p in model.parameters()])
        # first_param_mean_abs = list(model.parameters())[0].abs().mean().item()
        # print(f'Epoch {i+1}, Valid RMSEs: {valid_rmses}, first param mean abs: {first_param_mean_abs:g}, '
        #       f'grad nonzeros: {grad_nonzeros}')
        print('.', end='')
        for i in range(n_models):
            if valid_rmses[i] < best_valid_rmses[i]:
                best_valid_rmses[i] = valid_rmses[i]
                for p, best_p in zip(model.parameters(), best_model_params):
                    best_p[i] = p[i]

    print('', flush=True)

    with torch.no_grad():
        for p, best_p in zip(model.parameters(), best_model_params):
            p.set_(best_p)









import torch
import math
import numpy as np
from pathlib import Path
from typing import *

from . import custom_paths
from . import utils


def batch_randperm(n_batch: int, n: int, device: str = 'cpu') -> torch.Tensor:
    """
    Returns multiple random permutations.
    :param n_batch: Number of permutations.
    :param n: Length of permutations.
    :param device: PyTorch Device to put the permutations on.
    :return: Returns a torch.Tensor of integer type of shape [n_batch, n] containing the n_batch permutations.
    """
    # batched randperm:
    # https://discuss.pytorch.org/t/batched-shuffling-of-feature-vectors/30188/4
    # https://github.com/pytorch/pytorch/issues/42502
    return torch.stack([torch.randperm(n, device=device) for i in range(n_batch)], dim=0)


def seeded_randperm(n: int, device: str, seed: int) -> torch.Tensor:
    """
    Returns a random permutation according to a seed
    :param n: Length of the permutation.
    :param device: PyTorch device to put the permutation on
    :param seed: Seed for the random sampling.
    :return: Returns a torch.Tensor of shape [n] containing the permutation.
    """
    generator = torch.Generator()
    generator.manual_seed(seed)
    return torch.randperm(n, generator=generator).to(device)


class DictDataset:
    """
    Represents a data set containing multiple tensors that can be accessed by their name (a string),
    for example {'x': inputs, 'y': targets}.
    """
    def __init__(self, tensors: Dict[str, torch.Tensor], device: str = None):
        """
        :param tensors: Dictionary of names and tensors.
        All tensors should have the same shape[0], i.e., the same number of samples.
        All tensors should have two dimensions,
        i.e., scalar targets have to be passed with a second dimension of shape 1.
        :param device: PyTorch device that the tensors should be moved to.
        If device is None, all tensors are moved to the device of the first tensor.
        """
        self.device = device if device is not None else next(iter(tensors.values())).device
        self.n_samples = next(iter(tensors.values())).shape[0]
        self.tensors = None if tensors is None else {key: t.to(device) for key, t in tensors.items()}

    def get_batch(self, idxs: torch.Tensor) -> Dict[str, torch.Tensor]:
        """
        Returns the tensors corresponding to the batch indexed by idxs.
        :param idxs: Tensor of indices to index the tensors of this object with.
        :return: Returns a dictionary {key: t[idxs, :] for key, t in self.tensors.items()}
        """
        return {key: t[idxs, :] for key, t in self.tensors.items()}

    def get_sub_dataset(self, idxs: torch.Tensor) -> 'DictDataset':
        """
        Returns a data set representing the batch given by the indices idxs.
        :param idxs: Tensor of indices used to index the tensors of this object with.
        :return: Returns the DictDataset with the subset of samples specified by idxs.
        """
        return DictDataset(self.get_batch(idxs), device=self.device)

    def __len__(self) -> int:
        """
        :return: Returns the number of samples of the tensors.
        """
        return self.n_samples

    def to(self, device: str) -> 'DictDataset':
        """
        Move all tensors to the given device.
        :param device: PyTorch Device.
        :return: Returns a DictDataset with all tensors moved to the given device.
        """
        return DictDataset(self.tensors, device=device)


class ParallelDictDataLoader:
    """
    This class enables vectorized data loading from DictDatasets,
    i.e., if multiple models are trained in parallel on the same data set,
    this class selects batches independently for each of the models.
    If vectorization is not needed, it is of course possible to use this class with only one model.
    """
    def __init__(self, ds: DictDataset, idxs: torch.Tensor, batch_size: int, shuffle: bool = False,
                 adjust_bs: bool = True, drop_last: bool = False, output_device: Optional[str] = None):
        """
        :param dataset: A DictDataset from which tensors should be loaded
        :param idxs: Vectorized tensor of indices that specify a subset of ds from which data should be loaded.
        The tensor should have shape [n_parallel, n_samples],
        where n_parallel is the number of batches that should be drawn in a vectorized fashion,
        and n_samples is the number of samples that should be selected from ds.
        For example, if ensembling with 3 models should be used,
        ds represents the train+val+pool+test sets and idxs represents the indices belonging to the train set,
        then n_parallel=3 and n_samples=n_train.
        :param batch_size: default batch size, might be automatically adjusted
        :param shuffle: whether the dataset should be shuffled before each epoch
        :param adjust_bs: whether the batch_size may be lowered
        so that the batches are of more equal size while keeping the number of batches the same
        :param drop_last: whether the last batch should be omitted if it is smaller than the other ones
        :param output_device: The device that the returned data should be on
        (if None, take the device where the data already is)
        """
        self.ds = ds
        self.idxs = idxs.to(ds.device)
        self.n_parallel = idxs.shape[0]
        self.n_samples = idxs.shape[1]
        self.output_device = ds.device if output_device is None else output_device
        self.adjust_bs = adjust_bs
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.specified_batch_size = batch_size
        self.batch_size = min(batch_size, self.n_samples)

        if self.drop_last:
            self.n_batches = math.floor(self.n_samples / self.batch_size)
            if adjust_bs:
                self.batch_size = math.floor(self.n_samples / self.n_batches)
            self.sep_idxs = [self.batch_size * i for i in range(self.n_batches + 1)]
        else:
            self.n_batches = math.ceil(self.n_samples / self.batch_size)
            if adjust_bs:
                self.batch_size = math.ceil(self.n_samples / self.n_batches)
            self.sep_idxs = [self.batch_size * i for i in range(self.n_batches)] + [self.n_samples]

    def get_num_samples(self):
        """
        :return: Returns the number of samples that is sampled from (i.e. idxs.shape[1])
        """
        return self.n_samples

    def get_num_iterated_samples(self):
        """
        :return: Returns the number of samples that are visited in one epoch
        (might be less than get_num_samples() if drop_last=True).
        """
        if self.drop_last:
            return self.n_batches * self.batch_size
        return self.get_num_samples()

    def __len__(self):
        """
        :return: Returns the number of batches per epoch.
        """
        return self.n_batches

    def __iter__(self) -> Iterable[Dict[str, torch.Tensor]]:
        """
        Allows to iterate over batches of an epoch.
        :return: Returns an iterator that allows to iterate over dictionaries of the form {name: tensor}
        with tensor.shape[0] <= batch_size.
        """
        if self.shuffle:
            perms = batch_randperm(self.n_parallel, self.n_samples, device=self.ds.device)
            for start, stop in zip(self.sep_idxs[:-1], self.sep_idxs[1:]):
                batches = self.ds.get_batch(idxs=self.idxs.gather(1, perms[:, start:stop]))
                yield {key: t.to(self.output_device) for key, t in batches.items()}
        else:
            for start, stop in zip(self.sep_idxs[:-1], self.sep_idxs[1:]):
                batches = self.ds.get_batch(idxs=self.idxs[:, start:stop])
                yield {key: t.to(self.output_device) for key, t in batches.items()}


class DataInfo:
    """
    Represents information about a data set.
    """
    def __init__(self, ds_name: str, n_tvp: int, n_test: int, n_features: int,
                 train_test_split: Optional[int]):
        """
        :param ds_name: Name of the data set.
        :param n_tvp: Number of train+val+pool samples.
        :param n_test: Number of test samples.
        :param n_features: Number of input features of the data set.
        :param train_test_split: Set to None if the data set does not have a fixed (train+val+pool)-test split.
        If this is an int, it is interpreted such that the train+val+pool set are the first train_test_split samples
        and the test set are the remaining samples.
        """
        # tvp = train+val+pool
        self.ds_name = ds_name
        self.n_features = n_features
        self.n_tvp = n_tvp
        self.n_test = n_test
        self.n_samples = n_tvp + n_test
        self.train_test_split = train_test_split

    def save(self):
        """
        Saves this object to the path of the corresponding data set.
        """
        utils.serialize(Path(custom_paths.get_data_path()) / 'data' / self.ds_name / 'data_info.pkl', self)


class Task:
    """
    Represents a task, i.e., a data set and information what to do on the data set (how many batch AL steps etc).
    """
    def __init__(self, data_info: DataInfo, task_name: str, n_train: int, n_valid: int,
                 al_batch_sizes: List[int]):
        """
        Constructor. The actual data belonging to the task is loaded lazily, i.e., only when it is needed.
        :param data_info: DataInfo object representing the data set.
        :param task_name: Name of the task.
        This can be similar to the name of the data set
        but may include information about the number and size of batches.
        :param n_train: Number of initial training samples.
        :param n_valid: Number of validation samples.
        The remaining data_info.n_tvp - n_train - n_valid samples are used as initial pool samples.
        :param al_batch_sizes: List of batch sizes to acquire during batch active learning.
        """
        self.task_name = task_name
        self.data_info = data_info
        self.data = None  # load lazily
        self.n_train = n_train
        self.n_valid = n_valid
        self.n_pool = data_info.n_tvp - n_train - n_valid
        self.n_test = data_info.n_test
        self.al_batch_sizes = al_batch_sizes

    def get_data(self) -> DictDataset:
        """
        :return: Returns a DictDataset containing the data set, with names 'X' for the inputs and 'y' for the targets.
        The tensor shape belonging to 'X' is [n_samples, n_features]
        and the tensor shape belonging to 'y' is [n_samples, 1].
        """
        if self.data is None:
            base_path = Path(custom_paths.get_data_path()) / 'data' / self.data_info.ds_name
            X = np.load(f'{base_path}/X.npy')
            y = np.load(f'{base_path}/y.npy')
            self.data = DictDataset({'X': torch.as_tensor(X), 'y': torch.as_tensor(y)})
        return self.data

    @staticmethod
    def get_tabular_tasks(al_batch_sizes: List[int], n_train: int = 1024, n_valid: int = 1024,
                          ds_names: Optional[List[str]] = None, desc: Optional[str] = None) -> List['Task']:
        """
        :param al_batch_sizes: List of batch sizes specifying how many samples should be acquired
        in each batch active learning step.
        len(al_batch_sizes) is the number of batch active learning steps.
        :param n_train: Number of initial training samples.
        :param n_valid: Number of validation samples.
        :param ds_names: Names of data sets that should be used.
        If None is specified, all downloaded data sets are used.
        :param desc: Suffix that should be appended to the dataset names to obtain the task name.
        For example, if desc='256x16' and the current ds_name is 'sgemm',
        then the resulting task name is 'sgemm_256x16'. If desc is None, only the ds_name is used as the task name,
        without '_' at the end.
        :return: Returns a list of tasks, one task per data set.
        """
        base_path = Path(custom_paths.get_data_path())
        data_path = base_path / 'data'
        tasks = []
        if ds_names is None:
            ds_names = [dir.name for dir in data_path.iterdir()]
            ds_names.sort()
        for ds_name in ds_names:
            data_info = utils.deserialize(data_path / ds_name / 'data_info.pkl')
            tasks.append(Task(data_info, ds_name + ('' if desc is None else '_' + desc), n_train, n_valid,
                              al_batch_sizes))
        return tasks


class TaskSplit:
    """
    Represents one particular train-val-pool-test split of a task.
    It also preprocesses the task data according to the split.
    """
    def __init__(self, task: Task, id: int, use_pool_for_normalization: bool = False):
        """
        Creates the split and preprocesses the data. The preprocessed data set is stored in self.data.
        The idxs for train, val, pool, test can be found in
        self.train_idxs, self.valid_idxs, self.pool_idxs, self.test_idxs.
        :param task: Task to split.
        :param id: Identifier of the split. Also serves as a seed for creating the split.
        :param use_pool_for_normalization: Whether to compute the statistics for centering and standardization
        only on the (initial) train set or on train+pool sets. If the train set is small,
        """
        self.al_batch_sizes = task.al_batch_sizes
        self.data = task.get_data()
        self.id = id
        self.task_name = task.task_name
        self.n_samples = task.data_info.n_samples
        self.use_pool_for_normalization = use_pool_for_normalization

        old_state = np.random.get_state()
        np.random.seed(id)
        if task.data_info.train_test_split is not None:
            n_tvp = task.data_info.train_test_split
            tvp_perm = np.random.default_rng(id).permutation(n_tvp)
            self.test_idxs = np.arange(n_tvp, self.n_samples, dtype=np.int)
        else:
            perm = np.random.permutation(self.n_samples)
            n_tvp = self.n_samples - task.n_test
            tvp_perm = perm[:n_tvp]
            self.test_idxs = perm[n_tvp:]
        np.random.set_state(old_state)
        s1 = task.n_train
        s2 = s1 + task.n_valid
        self.train_idxs = tvp_perm[:s1]
        self.valid_idxs = tvp_perm[s1:s2]
        self.pool_idxs = tvp_perm[s2:]

        # preprocess tensors
        X = self.data.tensors['X']
        if use_pool_for_normalization:
            norm_idxs = np.concatenate([self.train_idxs, self.pool_idxs], axis=0)
        else:
            norm_idxs = self.train_idxs
        X_norm = X[norm_idxs]
        X = (X - X_norm.mean(dim=0, keepdim=True)) / (X_norm.std(dim=0, keepdim=True) + 1e-30)
        X = 5 * torch.tanh(0.2 * X)
        self.data = DictDataset({'X': X, 'y': self.data.tensors['y']})

    def get_data(self) -> DictDataset:
        """
        :return: Returns the DictDataset representing all of the data (train+val+pool+test)
        """
        return self.data

    def get_train_idxs(self) -> torch.Tensor:
        """
        :return: Returns the indices for the (initial) training data.
        """
        return self.train_idxs

    def get_valid_idxs(self) -> torch.Tensor:
        """
        :return: Returns the indices for the validation data.
        """
        return self.valid_idxs

    def get_pool_idxs(self) -> torch.Tensor:
        """
        :return: Returns the indices for the (initial) pool data.
        """
        return self.pool_idxs

    def get_test_idxs(self) -> torch.Tensor:
        """
        :return: Returns the indices for the test data.
        """
        return self.test_idxs



import sys
from .evaluation.plotting import *
from .evaluation.visualize_lcmd import create_lcmd_plots


def plot_all(results: ExperimentResults, alg_names: List[str], with_batch_size_plots: bool = True):
    selected_results = results.filter_alg_names(alg_names)
    literature_results = results.filter_alg_names(['NN_random', 'NN_maxdiag_ll_train', 'NN_maxdet-p_ll_train',
                                                   'NN_bait-fb-p_ll_train',
                                                   'NN_fw-p_ll_acs-rf-hyper-512', 'NN_maxdist-tp_ll',
                                                   'NN_kmeanspp-p_ll_train', 'NN_lcmd-tp_grad_rp-512'])
    literature_names = ['No BMAL', 'BALD', 'BatchBALD', 'BAIT', 'ACS-FW', 'Core-Set / FF-Active', 'BADGE', 'Ours']

    print('Generating tables...')
    save_latex_table_all_algs(results, 'table_all_algs.txt')
    save_latex_table_data_sets(selected_results, 'table_data_sets.txt')
    save_latex_table_data_sets(selected_results, 'table_data_sets_lasterror.txt', use_last_error=True)
    save_latex_table_data_sets(selected_results, 'table_data_sets_nolog.txt', use_log=False)
    save_latex_table_data_sets(selected_results, 'table_data_sets_nolog_lasterror.txt', use_log=False,
                               use_last_error=True)

    print('Creating learning curve plots...')
    plot_learning_curves_metrics_subplots(results=selected_results, filename='learning_curves_metrics.pdf')
    plot_multiple_learning_curves(results=selected_results, filename='learning_curves_rmse_maxe.pdf',
                                  metric_names=['rmse', 'maxe'])
    plot_multiple_learning_curves(results=selected_results, filename='learning_curves_q95_q99.pdf',
                                  metric_names=['q95', 'q99'])
    for metric_name in metric_names:
        plot_learning_curves(results=selected_results, filename=f'learning_curves_{metric_name}.pdf',
                             metric_name=metric_name)
        plot_learning_curves(results=literature_results, filename=f'learning_curves_literature_{metric_name}.pdf',
                             metric_name=metric_name, labels=literature_names, figsize=(6, 5))
        plot_learning_curves(results=literature_results, filename=f'learning_curves_literature_wide_{metric_name}.pdf',
                             metric_name=metric_name, labels=literature_names, figsize=(6, 3.5))

    print('Creating individual learning curve plots with subplots...')
    for metric_name in ['mae', 'rmse', 'q95', 'q99', 'maxe']:
        plot_learning_curves_individual_subplots(results=selected_results,
                                                 filename=f'learning_curves_individual_{metric_name}.pdf',
                                                 metric_name=metric_name)
    print('Creating error variation plots...')
    plot_error_variation(results, 'skewness_ri_lcmd-tp_grad_rp-512.pdf', metric_name='rmse', alg_name='NN_lcmd-tp_grad_rp-512',
                         use_relative_improvement=True)

    print('Creating correlation plots...')
    for metric_name in metric_names:
        plot_correlation_between_methods(results=selected_results,
                                         filename=f'correlation_between_methods_{metric_name}.pdf',
                                         metric_name=metric_name)
    print('Creating individual learning curve plots...')
    for metric_name in ['mae', 'rmse', 'q95', 'q99', 'maxe']:
        plot_learning_curves_individual(results=selected_results, metric_name=metric_name)

    if with_batch_size_plots:
        # batch size plots
        print('Creating batch size plots...')
        plot_batch_sizes_metrics_subplots(results=selected_results, filename='batch_sizes_metrics.pdf')
        plot_multiple_batch_sizes(results=selected_results, filename='batch_sizes_rmse_maxe.pdf',
                                  metric_names=['rmse', 'maxe'])
        plot_multiple_batch_sizes(results=selected_results, filename='batch_sizes_q95_q99.pdf',
                                  metric_names=['q95', 'q99'])
        for metric_name in metric_names:
            plot_batch_sizes(results=selected_results, filename=f'batch_sizes_{metric_name}.pdf',
                             metric_name=metric_name, figsize=(5, 5))
            plot_batch_sizes(results=selected_results, filename=f'batch_sizes_wide_{metric_name}.pdf',
                             metric_name=metric_name, figsize=(6, 3.5))
        print('Creating individual batch size plots with subplots...')
        for metric_name in ['mae', 'rmse', 'q95', 'q99', 'maxe']:
            plot_batch_sizes_individual_subplots(results=selected_results,
                                                 filename=f'batch_sizes_individual_{metric_name}.pdf',
                                                 metric_name=metric_name)
        print('Creating individual batch size plots...')
        for metric_name in ['mae', 'rmse', 'q95', 'q99', 'maxe']:
            plot_batch_sizes_individual(results=selected_results, metric_name=metric_name)


if __name__ == '__main__':
    metric_names = ['mae', 'rmse', 'q95', 'q99', 'maxe']
    if len(sys.argv) > 1:
        exp_names = [sys.argv[1]]
    else:
        available_names = utils.getSubfolderNames(custom_paths.get_results_path())
        exp_names = [name for name in available_names if name in ['relu', 'silu']]

    for exp_name in exp_names:
        print(f'----- Running evaluation for {exp_name} experiments -----')
        print('Loading experiment results...')
        results = ExperimentResults.load(exp_name)
        print('Loaded experiment results')
        print_avg_results(results)
        # print_all_task_results(results)
        print('Analyzing results')
        results.analyze_errors()
        results.analyze_eff_dims()

        if exp_name == 'relu':
            # selected algs for ReLU (best ones in terms of RMSE after ignoring slow ones, see table in the paper)
            alg_names_relu = ['NN_random', 'NN_maxdiag_grad_rp-512_acs-rf-512', 'NN_maxdet-p_grad_rp-512_train',
                              'NN_bait-f-p_grad_rp-512_train',
                              'NN_fw-p_grad_rp-512_acs-rf-hyper-512', 'NN_maxdist-p_grad_rp-512_train',
                              'NN_kmeanspp-p_grad_rp-512_acs-rf-512',
                              'NN_lcmd-tp_grad_rp-512']
            plot_all(results, alg_names=alg_names_relu)
        elif exp_name == 'silu':
            # selected algs for SiLU
            alg_names_silu = ['NN_random', 'NN_maxdiag_grad_rp-512_train', 'NN_maxdet-p_grad_rp-512_train',
                              'NN_bait-f-p_grad_rp-512_train',
                              'NN_fw-p_grad_rp-512_acs-rf-hyper-512', 'NN_maxdist-tp_grad_rp-512',
                              'NN_kmeanspp-tp_grad_rp-512',
                              'NN_lcmd-tp_grad_rp-512']
            plot_all(results, alg_names=alg_names_silu, with_batch_size_plots=False)

        print('Finished plotting')
        print()

    print('Creating lcmd visualization...')
    create_lcmd_plots(n_train=1, n_pool=500, n_steps=20)



import torch.distributions
import numpy as np
from .bmdal.layer_features import *


class SplittableModule(nn.Module):
    """
    Base class for vectorized (multiple NNs in parallel) training modules that can be split into non-vectorized modules
    """
    def get_single_model(self, i: int) -> nn.Module:
        raise NotImplementedError()


def get_act_layer(act='relu', **config):
    if act == 'relu':
        return nn.ReLU()
    if act == 'selu':
        return nn.SELU()
    if act == 'tanh':
        return nn.Tanh()
    if act == 'sigmoid':
        return nn.Sigmoid()
    if act == 'gelu':
        return nn.GELU()
    if act == 'relu6':
        return nn.ReLU6()
    if act == 'leaky_relu':
        return nn.LeakyReLU(negative_slope=config.get('lrelu_a', 1e-2))
    if act == 'rrelu':
        return nn.RReLU()
    if act == 'elu':
        return nn.ELU()
    if act == 'hardtanh':
        return nn.Hardtanh()
    if act == 'softplus':
        return nn.Softplus()
    if act == 'silu':
        return nn.SiLU()

    raise RuntimeError(f'Unknown activation "{act}"')


def get_parallel_act_layer(act='relu', **config):
    layer = get_act_layer(act, **config)
    return ParallelLayerWrapper(layer)


class ParallelLayerWrapper(SplittableModule):
    """
    Wraps a module without parameters into a SplittableModule.
    """
    def __init__(self, module: nn.Module):
        super().__init__()
        self.module = module

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.module(x)

    def get_single_model(self, i: int) -> nn.Module:
        return self


class ParallelLinearLayer(SplittableModule):
    """
    Linear layer, vectorized for training multiple NNs in parallel
    """
    def __init__(self, n_models, in_features, out_features, weight_gain=0.25, weight_init_mode='normal',
                 weight_init_gain=1.0, bias_gain=0.1, bias_init_gain=1.0, bias_init_mode='zero', **config):
        super().__init__()
        self.n_models = n_models
        self.in_features = in_features
        self.out_features = out_features

        self.weight = nn.Parameter(torch.Tensor(n_models, in_features, out_features))
        self.bias = nn.Parameter(torch.Tensor(n_models, out_features))

        with torch.no_grad():
            if weight_init_mode == 'normal':
                self.weight.normal_()
                self.weight *= weight_init_gain
            elif weight_init_mode == 'zero':
                self.weight.zero_()
            else:
                raise RuntimeError(f'Unknown weight_init_mode {weight_init_mode}')

            if bias_init_mode == 'normal':
                self.bias.normal_()
                self.bias *= bias_init_gain
            elif bias_init_mode == 'zero':
                self.bias.zero_()
            else:
                raise RuntimeError(f'Unknown weight_init_mode {weight_init_mode}')

        self.weight_factor = weight_gain/np.sqrt(in_features)
        self.bias_factor = bias_gain

    def get_single_model(self, i: int):
        return LinearLayer(weight=self.weight[i], bias=self.bias[i],
                           weight_factor=self.weight_factor, bias_factor=self.bias_factor)

    def forward(self, x):
        result = self.weight_factor * x.matmul(self.weight) + self.bias_factor * self.bias[:, None, :]
        return result


class LambdaLayer(SplittableModule):
    def __init__(self, f):
        super().__init__()
        self.f = f

    def get_single_model(self, i: int) -> nn.Module:
        return self

    def forward(self, x):
        return self.f(x)


class ParallelSequential(SplittableModule):
    def __init__(self, *layers):
        super().__init__()
        self.layers = nn.ModuleList(layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for l in self.layers:
            x = l(x)
        return x

    def get_single_model(self, i: int) -> nn.Module:
        return ParallelSequential(*[l.get_single_model(i) for l in self.layers])



from pathlib import Path

import matplotlib
#matplotlib.use('Agg')

matplotlib.use('pdf')
matplotlib.rcParams.update({
    "pgf.texsystem": "pdflatex",
    'font.family': 'serif',
    'font.size': 10,
    'text.usetex': True,
    'pgf.rcfonts': False,
    # 'legend.framealpha': 0.5,
    'text.latex.preamble': r'\usepackage{times} \usepackage{amsmath} \usepackage{amsfonts} \usepackage{amssymb}'
})
import matplotlib.pyplot as plt
import numpy as np

from .. import utils
from .. import custom_paths


def get_data(n_train, n_pool):
    n = n_train + n_pool
    x = []
    means = [[0, 0], [2, 0], [4, 0]]
    vars = [[0.5, 1], [2, 0.1], [0.5, 1]]
    probs = [0.7, 0.2, 0.1]
    for i in range(n):
        idx = np.random.choice([0, 1, 2], p=probs)
        x.append(np.asarray(means[idx]) + np.asarray(vars[idx]) * np.random.randn(2))
    x = np.asarray(x)
    return x[:n_train], x[n_train:]


def plot_step(x_train, x_pool, filename: str):
    fix, ax = plt.subplots(figsize=(3.5, 2.6))
    plt.tight_layout()
    sq_dists = np.linalg.norm(x_train[:, None, :] - x_pool[None, :, :], axis=-1)**2
    cluster_idxs = np.argmin(sq_dists, axis=0)
    min_sq_dists = np.min(sq_dists, axis=0)
    cluster_sizes = np.bincount(cluster_idxs, weights=min_sq_dists, minlength=x_train.shape[0])
    max_cluster_size = np.max(cluster_sizes)
    is_in_largest_cluster = cluster_sizes[cluster_idxs] == max_cluster_size
    min_sq_dists[~is_in_largest_cluster] = -np.Inf
    best_idx = np.argmax(min_sq_dists)
    for i in range(x_pool.shape[0]):
        color = '#FF8800' if is_in_largest_cluster[i] else '#4444FF'
        point = x_pool[i, :]
        center = x_train[cluster_idxs[i], :]
        plt.plot([center[0], point[0]], [center[1], point[1]], '-', color=color, alpha=0.2)
        # if i == best_idx:
        #     plt.plot([point[0]], [point[1]], '.', color='r')
    plt.plot(x_pool[:, 0], x_pool[:, 1], '.', color='#AAAAAA', alpha=0.6, markeredgewidth=0)
    plt.plot(x_pool[best_idx, 0], x_pool[best_idx, 1], '.', color='r', markeredgewidth=0)
    plt.plot(x_train[:, 0], x_train[:, 1], '.', color='k', markeredgewidth=0)
    ax.set_aspect('equal')
    plt.xlabel('$x_1$')
    plt.ylabel('$x_2$')
    plt.xticks([], [])
    plt.yticks([], [])
    plt.tight_layout()
    plt.savefig(filename, dpi=300)
    plt.close()
    return best_idx


def create_lcmd_plots(n_train, n_pool, n_steps=2, use_png=False):
    np.random.seed(0)
    x_train, x_pool = get_data(n_train, n_pool)
    for i in range(n_steps):
        plots_path = Path(custom_paths.get_plots_path())
        if use_png:
            filename = plots_path / 'lcmd_visualization_png' / f'lcmd_{n_train}_{n_pool}_{i}.png'
        else:
            filename = plots_path / 'lcmd_visualization' / f'lcmd_{n_train}_{n_pool}_{i}.pdf'
        utils.ensureDir(filename)
        best_idx = plot_step(x_train, x_pool, filename=str(filename))
        x_train = np.concatenate([x_train, x_pool[best_idx:best_idx+1, :]], axis=0)
        x_pool = np.delete(x_pool, best_idx, axis=0)




import pandas as pd
import numpy as np
import requests
import shutil
from typing import *
import openml
import mat4py

from . import custom_paths
from . import utils
from .data import DataInfo



def download_if_not_exists(url: str, dest: str):
    """
    Simple function for downloading a file from an url if no file at the destination path exists.
    :param url: URL of the file to download.
    :param dest: Path where to save the downloaded file.
    """
    # following https://dzone.com/articles/simple-examples-of-downloading-files-using-python
    utils.ensureDir(dest)
    if not utils.existsFile(dest):
        print('Downloading ' + url, flush=True)
        # file = requests.get(url)
        # open(dest, 'wb').write(file.content)
        r = requests.get(url, stream=True)
        with open(dest, 'wb') as f:
            print('Progress (dot = 1 MB): ', end='', flush=True)
            for ch in r.iter_content(chunk_size=1024**2):
                print('.', end='', flush=True)
                f.write(ch)
            print(flush=True)


class PandasTask:
    """
    This class represents a task (data set with indicated target variable) given by Pandas DataFrames.
    Additionally, a dedicated train-test split can be specified
    and the name of the data set needs to be specified for saving.
    This class provides a variety of methods for altering the task by different preprocessing methods.
    """
    def __init__(self, x_df: pd.DataFrame, y_df: pd.Series, ds_name: str, cat_indicator: Optional[List[bool]] = None,
                 train_test_split: Optional[int] = None):
        """
        :param x_df: DataFrame containing the inputs (covariates).
        :param y_df: pd.Series containing the targets.
        :param ds_name: Name for saving the data set.
        :param cat_indicator: Optional.
        One may specify a list of booleans which indicate whether each column of x is a category (True) or not (False).
        Otherwise, the column types in x_df will be used to decide whether a column is categorical or not.
        :param train_test_split: Optional. An integer can be specified as the index of the first test sample,
        if the data set has a dedicated test set part at the end.
        """
        if cat_indicator is None:
            cat_indicator = [not pd.api.types.is_numeric_dtype(x_df[x_df.columns[i]]) for i in range(len(x_df.columns))]
        else:
            # this is a fix coming from a different codebase
            # because category_indicator[0] was False for the dataset MIP-2016-regression
            # despite the column being categorical  (dtype=object)
            cat_indicator = [v or not pd.api.types.is_numeric_dtype(x_df[x_df.columns[i]])
                             for i, v in enumerate(cat_indicator)]
        if len(x_df.columns) != len(cat_indicator):
            raise ValueError('x.shape[1] != len(cat_indicator)')

        self.x_df = x_df  # should be (sparse) pd.DataFrame
        # should be (sparse) pd.Series  (i.e. a single column of a DataFrame)
        self.y_df = y_df
        self.ds_name = ds_name

        self.cat_cols = [x_df.columns[i] for i in range(len(x_df.columns)) if cat_indicator[i]]
        self.cont_cols = [x_df.columns[i] for i in range(len(x_df.columns)) if not cat_indicator[i]]
        self.train_test_split = train_test_split

    def get_n_samples(self):
        """
        :return: Returns the number of samples (number of rows in the DataFrame).
        """
        return len(self.x_df)

    def remove_missing_cont(self):
        """
        Removes rows with missing values in continuous columns.
        """
        print('removing columns with missing continuous values')
        if len(self.cont_cols) == 0:
            return  # no continuous columns

        not_nan_rows = self.x_df.notna().all(axis=1)
        self.x_df = self.x_df.loc[not_nan_rows, :]
        self.y_df = self.y_df.loc[not_nan_rows]

    def normalize_regression_y(self):
        """
        Centers and standardizes the target variable.
        """
        print('normalizing regression y')
        y_np = np.asarray(self.y_df)
        self.y_df.loc[:] = (y_np - np.mean(y_np)) / (np.std(y_np) + 1e-30)

    def subsample_dfs_(self, dfs: List[pd.DataFrame], max_n_samples: int) -> List[pd.DataFrame]:
        """
        Internal method for jointly subsampling multiple Pandas DataFrames of the same number of rows.
        :param dfs: Data Frames.
        :param max_n_samples: Maximum number of remaining rows.
        :return: Returns a List of potentially subsampled Pandas DataFrames.
        """
        if len(dfs[0]) <= max_n_samples:
            return dfs
        print(f'subsampling from {len(dfs[0])} samples to {max_n_samples}')
        idxs = np.random.default_rng(12345).permutation(len(dfs[0]))[:max_n_samples]
        return [df.iloc[idxs] for df in dfs]

    def subsample(self, max_tvp_samples: int, max_test_samples: int):
        """
        Subsamples the data set if necessary to not exceed a given maximum size.
        :param max_tvp_samples: Maximum number of train+val+pool samples.
        :param max_test_samples: Maximum number of test samples.
        """
        if self.train_test_split is not None:
            dfs_train = self.subsample_dfs_([self.x_df.loc[:self.train_test_split],
                                             self.y_df.loc[:self.train_test_split]], max_n_samples=max_tvp_samples)
            dfs_test = self.subsample_dfs_([self.x_df.loc[self.train_test_split:],
                                            self.y_df.loc[self.train_test_split:]], max_n_samples=max_test_samples)
            self.train_test_split = len(dfs_train[0])
            self.x_df = pd.concat([dfs_train[0], dfs_test[0]], axis=0)
            self.y_df = pd.concat([dfs_train[1], dfs_test[1]], axis=0)
        else:
            dfs = self.subsample_dfs_([self.x_df, self.y_df], max_n_samples=max_tvp_samples + max_test_samples)
            self.x_df, self.y_df = dfs[0], dfs[1]

    def remove_constant_columns(self):
        """
        Removes columns with only a single value (this could happen after removing NaN values).
        """
        # taken from https://stackoverflow.com/questions/20209600/pandas-dataframe-remove-constant-column
        non_constant_columns = (self.x_df != self.x_df.iloc[0]).any()
        print(f'removing constant columns')
        self.x_df = self.x_df.loc[:, non_constant_columns]
        self.cat_cols = [key for key in self.cat_cols if key in self.x_df.columns]
        self.cont_cols = [key for key in self.cont_cols if key in self.x_df.columns]

    def one_hot_encode(self, max_one_hot_columns: int):
        """
        Applies one-hot encoding to categorical columns.
        :param max_one_hot_columns: Maximal number of allowed one-hot encoded columns.
        If more one-hot encoded columns would be generated,
        the categorical columns with the largest number of categories are not one-hot encoded.
        """
        cat_cols_with_size = [(col_name, self.x_df.loc[:, col_name].nunique())
                              for i, col_name in enumerate(self.cat_cols)]
        if len(cat_cols_with_size) == 0:
            return  # nothing to encode
        print('one-hot encoding columns')
        cat_cols_with_size.sort(key=lambda t: t[1])  # sort by size of category
        max_cat_size = cat_cols_with_size[-1][1]
        new_col_sum = 0
        for key, sz in cat_cols_with_size:
            new_col_sum += sz
            if new_col_sum > max_one_hot_columns:
                max_cat_size = sz-1
                break

        new_cat_cols = []

        for key, sz in cat_cols_with_size:
            if sz <= max_cat_size:
                print(f'one-hot encoding column {key} with {sz} unique elements')
                # following https://stackoverflow.com/questions/37292872/how-can-i-one-hot-encode-in-python
                # https://queirozf.com/entries/one-hot-encoding-a-feature-on-a-pandas-dataframe-an-example
                col = self.x_df[key].astype('category')
                dummies = pd.get_dummies(col, prefix=f'{key}_onehot_', drop_first=(sz == 2), dummy_na=True,
                                         dtype=np.int32)
                self.cont_cols.extend(list(dummies.columns))
                self.x_df.drop([key], axis=1, inplace=True)
                self.x_df = pd.concat([self.x_df, dummies], axis=1)
            else:
                new_cat_cols.append(key)
                print(f'categorical column {key} with {sz} unique values is not one-hot encoded due to size constraints')

        self.cat_cols = new_cat_cols

    def save(self, n_test: int):
        """
        Saves the data set in the folder f'{custom_paths.get_data_path()}/data/{self.ds_name}'.
        :param n_test: Desired number of test samples.
        """
        folder = f'{custom_paths.get_data_path()}/data/{self.ds_name}'
        x_cont = np.array(self.x_df.reindex(columns=self.cont_cols), dtype=np.float32)
        y = np.array(self.y_df, dtype=np.float32)
        n_tvp = self.get_n_samples() - n_test
        data_info = DataInfo(ds_name=self.ds_name, n_tvp=n_tvp, n_test=n_test,
                             n_features=x_cont.shape[1],
                             train_test_split=self.train_test_split)

        utils.serialize(f'{folder}/data_info.pkl', data_info)
        np.save(f'{folder}/X.npy', x_cont)
        np.save(f'{folder}/y.npy', y[:, None])

        # ds = DictDataset({'x_cont': torch.as_tensor(x_cont), 'x_cat': torch.as_tensor(x_cat),
        #                   'y': torch.as_tensor(y[:, None])},
        #                  {'x_cont': TensorInfo(feat_shape=[x_cont.shape[-1]]),
        #                   'x_cat': TensorInfo(cat_sizes=cat_sizes),
        #                   'y': TensorInfo(cat_sizes=[self.get_n_classes()])})
        # task_info = TaskInfo.from_ds(task_desc, ds)
        # return Task(task_info, ds)

    @staticmethod
    def from_openml_task_id(task_id: int, ds_name: str):
        """
        Creates a PandasTask from an OpenML task.
        If more options are desired, we refer to PandasTask.from_openml_dataset_id() instead.
        :param task_id: OpenML task id.
        :param ds_name: Short name of the data set for saving.
        :return: Returns a PandasTask representing the OpenML task.
        """
        task = openml.tasks.get_task(task_id, download_data=False)
        return PandasTask.from_openml_dataset_id(task.dataset_id, ds_name, task.target_name)

    @staticmethod
    def from_openml_dataset_id(dataset_id: int, ds_name: str, target: str,
                               ignore_columns: Optional[List[str]] = None,
                               use_log_target: bool = False):
        """
        Creates a PandasTask from an OpenML data set.
        :param dataset_id: OpenML data set id.
        :param ds_name: Short name of the data set for saving.
        :param target: Name of the target variable.
        :param ignore_columns: Optional. List of columns that should be removed.
        :param use_log_target: Whether the logarithm should be applied to the target column.
        :return: Returns a PandasTask representing the corresponding OpenML data set.
        """
        print(f'Importing dataset {ds_name}')
        openml.config.set_cache_directory('./openml_cache')
        dataset = openml.datasets.get_dataset(dataset_id, download_data=False)
        print(f'dataset name: {dataset.name}')
        # print(dataset.get_data(dataset_format='dataframe'))
        x_df, y_df, cat_indicator, names = dataset.get_data(target=target,
                                                            dataset_format='dataframe')

        if ignore_columns is not None:
            cat_indicator = [value for col_name, value in zip(x_df.columns, cat_indicator)
                                  if col_name not in ignore_columns]

        for key in ignore_columns or []:
            x_df.drop([key], axis=1, inplace=True)

        if use_log_target:
            y_df = np.log(y_df)

        print('Imported x_df:\n', x_df)
        print('Imported y_df:\n', y_df)

        return PandasTask(x_df, y_df, ds_name, cat_indicator)

    @staticmethod
    def from_uci(url: str, ds_name: str, zip_name: str, csv_name: str, target_col_idxs: List[int],
                 ignore_col_idxs: Optional[List[int]] = None, use_log_target: bool = False,
                 train_test_boundary: Optional[int] = None, has_header: bool = True,
                 continuous_nan_columns: Optional[List[int]] = None,
                 convert_to_cat_columns: Optional[List[int]] = None,
                 ignore_other_csv_files: bool = False, separator: Optional[str] = None):
        """
        Create a PandasTask object from a data set on the UCI repository.
        :param url: URL of the data set file.
        :param ds_name: Short name of the data set used for saving the data set.
        :param zip_name: Target name of the downloaded file.
        If the downloaded file is not zip/compressed, i.e., no unzipping is needed,
        zip_name should be the same as csv_name.
        Otherwise the file with name zip_name will be unzipped to the file with name csv_name.
        :param csv_name: Target name of the uncompressed file, see zip_name.
        :param target_col_idxs: List of indexes of target columns. Mostly, this will only have one element.
        :param ignore_col_idxs: List of indexes of columns to be removed.
        :param use_log_target: Whether the logarithm should be applied to the target value.
        :param train_test_boundary: Index of the first test sample.
        If (as in most cases) there is no dedicated test set, None should be specified.
        :param has_header: Whether the downloaded (unzipped) csv file has a row with column names that should be removed.
        :param continuous_nan_columns: Optional. List of indexes of numeric columns that can contain NaN values.
        This triggers a corresponding Pandas column conversion.
        :param convert_to_cat_columns: Optional.
        List of indexes of columns that should be converted to categorical type.
        :param ignore_other_csv_files: If set to True,
        unzipped csv files with file names other than csv_name will be ignored.
        If set to False, unzipped csv files with file names other than csv_name
        will be interpreted as a dedicated test set,
        and the file with name csv_name will be interpreted as the train+val+pool set.
        :param separator: Separator in the csv file. Default is ','.
        If the file with name csv_name is a tsv file, separator='\t' should be specified.
        :return: Returns a PandasTask object.
        """
        print(f'Importing dataset {ds_name}')
        base_path = custom_paths.get_data_path()
        raw_data_folder = f'{base_path}/raw_data/{ds_name}'
        zip_file = f'{raw_data_folder}/{zip_name}'
        csv_file = f'{raw_data_folder}/{csv_name}'
        download_if_not_exists(url, zip_file)
        if not utils.existsFile(csv_file):
            print('Unpacking zip file...')
            shutil.unpack_archive(zip_file, raw_data_folder)

        if separator is None:
            separator = ','

        print('Processing csv data...')
        if ignore_other_csv_files:
            non_train_files = []
        else:
            non_train_files = [file for file in utils.matchFiles(f'{raw_data_folder}/*.csv') if file != csv_file]
        df = pd.read_csv(csv_file, header='infer' if has_header else None, sep=separator)
        if len(non_train_files) > 0:
            train_test_boundary = len(df)
            df = pd.concat([df] + [pd.read_csv(file, header='infer' if has_header else None, sep=separator)
                                     for file in non_train_files])

        if continuous_nan_columns is not None:
            for col_idx in continuous_nan_columns:
                df.iloc[:, col_idx] = pd.to_numeric(df.iloc[:, col_idx], errors='coerce')

        if convert_to_cat_columns is not None:
            for col_idx in convert_to_cat_columns:
                df.iloc[:, col_idx] = df.iloc[:, col_idx].astype('category')

        input_column_names = []
        target_columns = []
        for i in range(len(df.columns)):
            if i in target_col_idxs:
                target_columns.append(df.iloc[:, i].to_numpy().astype(np.float32))
            elif ignore_col_idxs is None or i not in ignore_col_idxs:
                input_column_names.append(df.columns[i])

        y = np.median(np.stack(target_columns, axis=1), axis=1)
        if use_log_target:
            y = np.log(y)

        # https://stackoverflow.com/questions/20490274/how-to-reset-index-in-a-pandas-dataframe
        x_df = df.reindex(columns=input_column_names).reset_index(drop=True)
        y_df = pd.DataFrame({'y': y})['y']

        return PandasTask(x_df, y_df, ds_name,
                          train_test_split=train_test_boundary)


class PandasTaskPreprocessor:
    """
    This class allows to preprocess data sets given by PandasTask objects. Various options can be configured.
    """
    def __init__(self, min_n_samples: int, max_tvp_samples: int, max_test_samples: int, max_one_hot_columns: int):
        """
        :param min_n_samples: Minimum number of samples that a task must have
        after removing missing values in continuous columns.
        :param max_tvp_samples: Maximum number of samples for the train+val+pool sets.
        :param max_test_samples: Maximum number of samples for the test set.
        :param max_one_hot_columns: Maximum number of one-hot encoded columns that are allowed.
        If more would be generated, the categorical variables with the largest category sizes are removed.
        """
        self.min_n_samples = min_n_samples
        self.max_tvp_samples = max_tvp_samples
        self.max_test_samples = max_test_samples
        self.max_one_hot_columns = max_one_hot_columns

    def apply(self, pd_task: PandasTask):
        """
        Apply preprocessing to a PandasTask with the options given in the constructor
         and save the preprocessed data set under the name specified in the PandasTask.
        :param pd_task: PandasTask object holding infomation about the unprocessed data set.
        """
        pd_task.remove_missing_cont()
        if pd_task.get_n_samples() < self.min_n_samples:
            print(f'Task {pd_task.ds_name} has only {pd_task.get_n_samples()} samples after removing missing values, '
                  f'hence we discard it')
            return
        pd_task.subsample(max_tvp_samples=self.max_tvp_samples, max_test_samples=self.max_test_samples)
        pd_task.remove_constant_columns()
        pd_task.one_hot_encode(self.max_one_hot_columns)
        pd_task.normalize_regression_y()
        n_samples = pd_task.get_n_samples()
        if pd_task.train_test_split is not None:
            n_test = n_samples - pd_task.train_test_split
        else:
            n_test = max(int(0.2 * n_samples), n_samples - self.max_tvp_samples)
        pd_task.save(n_test)


def get_sarcos_pandas_task() -> PandasTask:
    """
    This is a separate function for downloading the sarcos data set, since it is not in the UCI / OpenML repositories.
    """
    print(f'Importing dataset sarcos')
    base_path = custom_paths.get_data_path()
    raw_data_folder = f'{base_path}/raw_data/sarcos'
    file_path = f'{raw_data_folder}/sarcos_inv.mat'
    download_if_not_exists('http://www.gaussianprocess.org/gpml/data/sarcos_inv.mat', file_path)
    # don't download test data since test_data = train_data[::10]
    # download_if_not_exists('http://www.gaussianprocess.org/gpml/data/sarcos_inv_test.mat',
    #                        f'{raw_data_folder}/sarcos_inv_test.mat')
    data = np.array(mat4py.loadmat(file_path)['sarcos_inv'])
    return PandasTask(x_df=pd.DataFrame(data[:, :-7]), y_df=pd.Series(data[:, -7]), ds_name='sarcos')


def import_all():
    proc = PandasTaskPreprocessor(min_n_samples=30000, max_tvp_samples=200000, max_test_samples=300000,
                                  max_one_hot_columns=300)

    proc.apply(PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00440/sgemm_product_dataset.zip',
                   ds_name='sgemm', zip_name='sgemm_product_dataset.zip', csv_name='sgemm_product.csv',
                   target_col_idxs=[14, 15, 16, 17], use_log_target=True))
    proc.apply(PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00206/slice_localization_data.zip',
                   ds_name='ct', zip_name='slice_localization_data.zip', csv_name='slice_localization_data.csv',
                   target_col_idxs=[385], ignore_col_idxs=[0]))
    proc.apply(PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00221/Reaction%20Network%20(Undirected).data',
                   ds_name='kegg_undir_uci', zip_name='kegg_undir_uci.csv', csv_name='kegg_undir_uci.csv',
                   target_col_idxs=[26], ignore_col_idxs=[0], has_header=False, continuous_nan_columns=[4]))
    # only use the Sydney part of the data set (could as well have used Adelaide, Perth or Tasmania)
    proc.apply(PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00494/WECs_DataSet.zip',
                   ds_name='wecs', zip_name='WECs_DataSet.zip', csv_name='WECs_DataSet/Sydney_Data.csv',
                   target_col_idxs=[48], has_header=False, ignore_other_csv_files=True))
    proc.apply(PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00335/online_video_dataset.zip',
                                   ds_name='online_video', zip_name='online_video_dataset.zip',
                                   csv_name='transcoding_mesurment.tsv', separator='\t', ignore_col_idxs=[0, 20],
                                   target_col_idxs=[21], has_header=True, ignore_other_csv_files=True))
    proc.apply(
        PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00493/datasets.zip',
                            ds_name='query_agg_count', zip_name='datasets.zip',
                            csv_name='Datasets/Range-Queries-Aggregates.csv', ignore_col_idxs=[0, 6, 7],
                            target_col_idxs=[5], has_header=True, ignore_other_csv_files=True))
    proc.apply(
        PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00246/3D_spatial_network.txt',
                            ds_name='road_network', zip_name='3D_spatial_network.txt',
                            csv_name='3D_spatial_network.txt', ignore_col_idxs=[0],
                            target_col_idxs=[3], has_header=False, ignore_other_csv_files=True))
    # note: we use only the testing data here with random splits since the training data is so small
    # and the testing data so large
    proc.apply(
        PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/poker/poker-hand-testing.data',
                            ds_name='poker', zip_name='data.csv', csv_name='data.csv',
                            convert_to_cat_columns=list(range(10)),
                            target_col_idxs=[10], has_header=False, ignore_other_csv_files=True))
    proc.apply(get_sarcos_pandas_task())
    proc.apply(PandasTask.from_openml_dataset_id(dataset_id=1200, ds_name='stock', target='company10'))
    proc.apply(PandasTask.from_openml_dataset_id(dataset_id=42454, ds_name='mlr_knn_rng', target='perf.logloss',
                                                 ignore_columns=['perf.mmce']))
    proc.apply(PandasTask.from_openml_dataset_id(dataset_id=42701, ds_name='methane', target='MM264',
                                                 ignore_columns=['MM263', 'MM256']))
    proc.apply(PandasTask.from_openml_dataset_id(dataset_id=42225, ds_name='diamonds', target='price'))
    proc.apply(PandasTask.from_openml_dataset_id(dataset_id=564, ds_name='fried', target='Y'))
    proc.apply(PandasTask.from_openml_dataset_id(dataset_id=42903, ds_name='protein', target='RMSD'))

    # ------ data sets that were discarded due to too many missing values or too low learnability ------
    # proc.apply(PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00304/BlogFeedback.zip',
    #                ds_name='blog', zip_name='BlogFeedback.zip', csv_name='blogData_train.csv',
    #                target_col_idxs=[280], has_header=False))
    # proc.apply(PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00332/OnlineNewsPopularity.zip',
    #                ds_name='news', zip_name='OnlineNewsPopularity.zip',
    #                csv_name='OnlineNewsPopularity/OnlineNewsPopularity.csv',
    #                target_col_idxs=[60], ignore_col_idxs=[0, 1]))
    # proc.apply(PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00203/YearPredictionMSD.txt.zip',
    #                ds_name='year', zip_name='YearPredictionMSD.txt.zip', csv_name='YearPredictionMSD.txt',
    #                target_col_idxs=[0], train_test_boundary=463715, has_header=False))
    # # date-time currently not handled
    # proc.apply(PandasTask.from_openml_dataset_id(dataset_id=42998, ds_name='metro', target='traffic_volume'))
    # proc.apply(PandasTask.from_openml_dataset_id(dataset_id=42163, ds_name='procurement', target='award_value_euro'))
    # # ignore V1, V5 is numerical but with missing value?, V27 might not be the true target
    # # but use kegg_undir_uci anyway
    # proc.apply(PandasTask.from_openml_dataset_id(dataset_id=4533, ds_name='kegg_undir', target='V27'))
    # proc.apply(PandasTask.from_openml_dataset_id(dataset_id=41540, ds_name='black_friday', target='Purchase'))
    # proc.apply(PandasTask.from_openml_dataset_id(dataset_id=4549, ds_name='buzz_twitter', target='Annotation'))
    # proc.apply(PandasTask.from_openml_dataset_id(dataset_id=1204, ds_name='bng_wine', target='quality'))
    # proc.apply(PandasTask.from_openml_dataset_id(dataset_id=1197, ds_name='2dplanes', target='y'))
    # proc.apply(PandasTask.from_openml_dataset_id(dataset_id=1213, ds_name='mv', target='y'))
    # proc.apply(PandasTask.from_openml_dataset_id(dataset_id=42705, ds_name='yolanda', target='101'))
    # proc.apply(PandasTask.from_openml_dataset_id(dataset_id=1203, ds_name='pwlinear', target='class'))

    # ------ kegg was discarded since we already have kegg_undir_uci, which has better learnability ------
    # proc.apply(PandasTask.from_uci('https://archive.ics.uci.edu/ml/machine-learning-databases/00220/Relation%20Network%20(Directed).data',
    #                ds_name='kegg', zip_name='kegg.csv', csv_name='kegg.csv',
    #                target_col_idxs=[23], ignore_col_idxs=[0], has_header=False))


if __name__ == '__main__':
    import_all()
    pass






